<?php

namespace Bazookas\AdminBundle\Controller\Base;

use Bazookas\AdminBundle\AdminElements\Interfaces\AdminElementInterface;
use Symfony\Bundle\FrameworkBundle\Controller\Controller;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;

abstract class BaseAdminController extends Controller
{
  //template layout constants
  const TEMPLATE_LAYOUT_BASIC = 'BazookasAdminBundle:AdminElements\Layout:defaultLayout.html.twig';
  const TEMPLATE_LAYOUT_CONTENT_ONLY = 'BazookasAdminBundle:AdminElements\Layout:contentOnlyLayout.html.twig';
  const TEMPLATE_LAYOUT_MENU_LESS = 'BazookasAdminBundle:AdminElements\Layout:menuLessLayout.html.twig';

  /**
   * The basic layout the page should use
   * @var string
   */
  public $layoutTemplate;

  /**
   * The elements the page contains
   * @var AdminElementInterface[]
   */
  public $elements;

  /**
   * The title of the page
   * @var string
   */
  public $title;

  /**
   * add custom styles
   * @var array
   */
  protected $styles = [];

  /**
   * add custom scripts
   * @var array
   */
  protected $scripts = [];

  public function __construct() {
    $this->elements = [];
    $this->layoutTemplate = self::TEMPLATE_LAYOUT_BASIC;
  }

  /**
   * Get the current layout template
   * @return string
   */
  public function getLayoutTemplate()
  {
    return $this->layoutTemplate;
  }

  /**
   * Set the current layout template
   * @param string $layoutTemplate
   * @return $this
   */
  public function setLayoutTemplate($layoutTemplate)
  {
    $this->layoutTemplate = $layoutTemplate;
    return $this;
  }

  /**
   * Get the array of AdminElementInterface elements that will be used to build this page
   * @return array|AdminElementInterface[]
   */
  public function getElements()
  {
    return $this->elements;
  }

  /**
   * Set the array of AdminElementInterface elements that will be used to build this page
   * @param array $elements
   * @return $this
   */
  public function setElements(array $elements)
  {
    $this->elements = $elements;

    return $this;
  }

  /**
   * Add a single AdminElementInterface element
   * @param AdminElementInterface $element
   * @return $this
   */
  public function addElement(AdminElementInterface $element) {
    if (!in_array($element, $this->elements)) {
      $this->elements[] = $element;
    }

    return $this;
  }

  /**
   * Remove a single AdminElementInterface element
   * @param AdminElementInterface $element
   * @return $this
   */
  public function removeElement(AdminElementInterface $element) {
    if (in_array($element, $this->elements)) {
      $index = array_keys($this->elements, $element);
      unset($this->elements[$index[0]]);
    }

    return $this;
  }

  /**
   * Get the page title
   * @return string
   */
  public function getTitle()
  {
    return $this->title;
  }

  /**
   * Set the page title
   * @param string $title
   * @return $this
   */
  public function setTitle($title)
  {
    $this->title = $title;

    return $this;
  }

  /**
   * Get the array of css stylesheets that should be included on this page
   * @return array
   */
  public function getStyles()
  {
    return $this->styles;
  }

  /**
   * Set the array of css stylesheets that should be included on this page
   * @param $styles
   * @return $this
   */
  public function setStyles(array $styles)
  {
    $this->styles = $styles;

    return $this;
  }

  /**
   * Get the array of javascript files taht will be included on this page
   * @return array
   */
  public function getScripts()
  {
    return $this->scripts;
  }

  /**
   * Set the array of javascript files taht will be included on this page
   * @param array $scripts
   * @return $this
   */
  public function setScripts(array $scripts)
  {
    $this->scripts = $scripts;

    return $this;
  }

  /**
   * @param Request $request
   * @param array $templateData
   * @param string $action
   * @return \Symfony\Component\HttpFoundation\Response
   */
  public function createResponse(Request $request, array $templateData, $action) {
    return $this->render(
      $this->getLayoutTemplate(), [
      'elements' => $this->getElements(),
      'templateData' => $templateData,
      'styles' => $this->getStyles(),
      'scripts' => $this->getScripts(),
    ]);
  }

  /**
   * Get the template data for the current request and action. Overwrite this as needed.
   * @param Request $request
   * @param string $action
   * @param array $data
   * @return array
   */
  protected function getTemplateData(Request $request, $action, $data = []) {
    return array_merge([
      'titleTranslationKey' => $this->getTitle()
    ], $data);
  }

  /**
   * Create an access denied redirect response
   * @param Request $request
   * @param string|null $route
   * @param null $routeParams
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   */
  protected function accessDeniedRedirect(Request $request, $route = null, $routeParams = null) {
    $translator = $this->get('translator');
    $this->addFlash(
      'error',
      $translator->trans('admin.list.messages.accessDenied', [], 'admin')
    );

    //send the user back where they belong!
    if (empty($route)) {
      $route = $request->get('_route');
    }
    if (is_null($routeParams)) {
      $routeParams = $request->get('_route_params');
    }

    if (empty($route)) {
      throw new AccessDeniedException();
    }

    return $this->redirectToRoute($route, $routeParams);
  }

  /**
   * Check whether the current logged in user has access to perform the action
   * @param string $action
   * @return mixed
   */
  protected abstract function hasAccess($action);

}
