<?php

namespace Bazookas\AdminBundle\Controller;

use Bazookas\AdminBundle\Filter\FilterService;
use Bazookas\AdminBundle\Util\PageAction;
use Bazookas\CommonBundle\Exception\LinkedEntityException;
use Doctrine\Common\Persistence\ObjectRepository;
use Doctrine\DBAL\Query\QueryBuilder;
use Doctrine\ORM\Mapping\ClassMetadata;
use Doctrine\ORM\Tools\Pagination\Paginator;
use Symfony\Bundle\FrameworkBundle\Controller\Controller;
use Symfony\Component\EventDispatcher\EventDispatcher;
use Symfony\Component\Form\Form;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Request;
use Bazookas\AdminBundle\Util\ListConfigurator;
use Bazookas\AdminBundle\Event\AdminListEvent;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;
use Bazookas\CommonBundle\Entity\Interfaces\EntityInterface;
use Bazookas\AdminBundle\Exception\Interfaces\AdminExceptionInterface;

/**
 * Extend this class to create admin lists
 * See UserAdminController as an example
 * TODO create a data repository/service to clean up this class a little bit :)
 *
 * NOTE: the entity's "canDelete" method is called upon deleting, so should be sufficient?
 *    - can be due to them not having the necessary access rights
 *    - can be due to the entity being linked to other entities
 */
class ListBaseController extends Controller
{
  const ACTION_LIST = 'list';
  const ACTION_EDIT = 'edit';
  const ACTION_ADD = 'add';
  const ACTION_VIEW = 'view';
  const ACTION_REMOVE = 'remove';
  const ACTION_EXPORT = 'export';
  const AVAILABLE_ACTIONS = [
    self::ACTION_LIST,
    self::ACTION_EDIT,
    self::ACTION_ADD,
    self::ACTION_VIEW,
    self::ACTION_REMOVE,
    self::ACTION_EXPORT
  ];

  /**
   * @var ListConfigurator
   */
  protected $config;

  protected $exportConfig;

  /**
   * @var string
   */
  protected $action;
  /**
   * @var int
   */
  protected $id;
  /**
   * @var EventDispatcher
   */
  protected $dispatcher;
  /**
   * @var AdminListEvent
   */
  protected $event;
  /**
   * @var ClassMetadata
   */
  protected $entityMetaData;

  /**
   * @var mixed anything custom you want to pass to the view, to be used in extending controllers
   * @deprecated just use the mergeTemplateData function below! :O
   **/
  protected $customViewData;

  /**
   * A service for handling filtering of a list view
   * @var FilterService
   */
  protected $filterService;

  /**
   * the constructor will create a new config instance and set some default templates
   * NOTE: it's important to call parent constructor when extending (or manually create config instance)
   **/
  function __construct() {
    // create a new config instance and define a local var for easy referencing during construct
    $config = $this->config = new ListConfigurator();

    $config->listTemplate = 'BazookasAdminBundle:List:index.html.twig';
    $config->editTemplate = 'BazookasAdminBundle:List:form.html.twig';
    $config->addTemplate = 'BazookasAdminBundle:List:form.html.twig';
    $config->viewTemplate = 'BazookasAdminBundle:List:dumpEntity.html.twig';

    //instantiate services
    $this->filterService = new FilterService($this->config);
  }

  /**
   * indexAction processes all requests and calls the correct action function
   * this way the routes can be defined in a single item
   * @param Request $request
   * @param string $_action
   * @param string $_id
   * @return Response
   **/
  public function indexAction(Request $request, $_action = null, $_id = null)
  {
    if (empty($_action)) {
      $_action = self::ACTION_LIST;
    }

    $this->action = $_action;
    $this->id = $_id;
    $this->entityMetaData = $this
      ->getDoctrine()
      ->getManager()
      ->getClassMetaData($this->config->entity)
    ;

    // create an event dispatcher
    $this->dispatcher = $this->get('event_dispatcher');
    $this->event = new AdminListEvent($_action, $_id, $this->config, $this->entityMetaData, $request);
    $this->addListeners($this->dispatcher);

    // before moving on: check access
    if (!$this->checkAccess($_action, $_id) || !$this->config->getAccess($_action)) {
      return $this->accessDeniedRedirect($request);
    }

    switch ($_action) {
      case self::ACTION_ADD:
        $this->createAddActions($request);
        $response = $this->addAction($request);
        break;
      case self::ACTION_EDIT:
        $this->createEditActions($request, $_id);
        $response = $this->editAction($request);
        break;
      case self::ACTION_VIEW:
        $this->createViewActions($request, $_id);
        $response = $this->viewAction($request);
        break;
      case self::ACTION_REMOVE:
        $response = $this->removeAction($request);
        break;
      case self::ACTION_LIST:
        $this->filterService->parseFilters($request);
        $this->filterService->parseOrder($request);
        $page = $request->query->get('page');
        if (!empty($page)) {
          $this->config->currentPage = intval($page);
        }
        $this->createListActions($request);
        $response = $this->listAction($request);
        break;
      case self::ACTION_EXPORT:
        $this->filterService->parseFilters($request);
        $this->filterService->parseOrder($request);
        $response = $this->exportAction($request);
        break;
      default:
        $response = $this->executeCustomActions($request, $_action, $_id);
    }

    return $response;
  }

  /**
   * Override this function to execute custom actions on this list base controller
   * @param Request $request
   * @param string $_action
   */
  protected function executeCustomActions(Request $request, $_action, $_id) {
    throw $this->createNotFoundException("No action '$_action' for this admin list");
  }

  /**
   * Override this function to add listeners to the current dispatcher
   * Only needed for controller specific listeners, global listeners can be added through services
   * @param dispatcher
   **/
  protected function addListeners($dispatcher) {
    // OVERRIDE ME
  }

  /**
   * List all entities
   * @param Request $request
   * @return Response
   **/
  public function listAction(Request $request) {
    $queryBuilder = $this->getQueryBuilder()
      // paging
      ->setFirstResult( max(0, ($this->config->currentPage - 1) * $this->config->limit) )
      ->setMaxResults( $this->config->limit )
    ;

    $this->filterService->executeOrder($queryBuilder);
    $this->filterService->executeFilters($queryBuilder);

    // dispatch a before list query event for other bundles
    // use this query builder in event listeners to provide app specific filters
    $this->event->queryBuilder = $queryBuilder;
    $this->dispatcher->dispatch(AdminListEvent::BEFORE_LIST_QUERY, $this->event);

    // get the final query
    $entities = $queryBuilder->getQuery()->getResult();

    // count the total rows
    $paginator = new Paginator($queryBuilder->getQuery(), true);
    $totalCount = count($paginator);

    $this->config->totalPages = intval(ceil($totalCount / $this->config->limit));

    $this->event->queryResult = $entities;
    $this->dispatcher->dispatch(AdminListEvent::AFTER_LIST_QUERY, $this->event);

    $templateData = array(
      'entities' => $entities,
      'titleTranslationKey' => $this->config->getTitle($this->action),
      'titleTranslationVariables' => array('%entity%' => $this->getEntityPluralName()),
      'showSearch' => $this->config->showSearch,
      'showFilter' => $this->config->showFilter,
      'filterForm' => $this->filterService->createFilterForm($this->get('form.factory')),
      'isOrderable' => ($this->config->isOrderable() && empty($this->filters)),
    );
    $templateData = $this->mergeTemplateData($templateData, self::ACTION_LIST);

    return $this->render($this->config->listTemplate, $templateData);
  }

  /**
   * add an entity using a formbuilder class
   * @param Request $request
   * @return Response
   **/
  public function addAction(Request $request) {
    //Execute access control check to make sure the user is allowed to perform the action
    $entity = new $this->entityMetaData->name();
    $auth = $this->get('security.authorization_checker');
    if (empty($entity) || !$entity->canAdd($auth, $this->getUser())) {
      return $this->accessDeniedRedirect($request);
    }

    $form = $this->createForm($this->config->getAddForm(), $entity);
    return $this->handleForm(
      $request,
      $form,
      $entity,
      $this->config->addTemplate,
      $this->config->getTitle($this->action),
      $this->config->addedMessage
    );
  }

  /**
   * edit an entity using a formbuilder class
   * @param Request $request
   * @return Response
   **/
  public function editAction(Request $request) {
    //Execute access control check to make sure the user is allowed to perform the action
    $entity = $this->getRepository()->find($this->id);
    $auth = $this->get('security.authorization_checker');

    if (empty($entity) || !$entity->canEdit($auth, $this->getUser())) {
      return $this->accessDeniedRedirect($request);
    }

    $this->event->queryResult = $entity;
    $this->event->entity = $entity;
    $this->dispatcher->dispatch(AdminListEvent::BEFORE_EDIT, $this->event);

    $form = $this->createForm($this->config->getEditForm(), $entity);
    return $this->handleForm(
      $request,
      $form,
      $entity,
      $this->config->editTemplate,
      $this->config->getTitle($this->action),
      $this->config->editedMessage
    );
  }

  /**
   * view an entity
   * @param Request $request
   * @return Response
   **/
  public function viewAction(Request $request) {
    //Execute access control check to make sure the user is allowed to perform the action
    $entity = $this->getRepository()->find($this->id);
    $auth = $this->get('security.authorization_checker');

    if (empty($entity) || !$entity->canView($auth, $this->getUser())) {
      return $this->accessDeniedRedirect($request);
    }

    $templateData = [
      'entity' => $entity,
      'titleTranslationKey' => $this->config->getTitle($this->action),
      'titleTranslationVariables' => array('%entity%' => $this->getEntityName()),
    ];
    $templateData = $this->mergeTemplateData($templateData, self::ACTION_VIEW);

    return $this->render($this->config->viewTemplate, $templateData);
  }

  public function removeAction(Request $request) {
    //Execute access control check to make sure the user is allowed to perform the action
    $entity = $this->getRepository()->find($this->id);
    $auth = $this->get('security.authorization_checker');
    if (empty($entity) || !$entity->canDelete($auth, $this->getUser())) {
      return $this->accessDeniedRedirect($request);
    }

    $translator = $this->get('translator');
    try {
      $this->removeEntity($this->id);

      // set flash message
      $this->addFlash(
        'info',
        $translator->trans('admin.list.messages.removed', array(
          '%entity%' => $this->getEntityName()
        ),'admin')
      );
    } catch(LinkedEntityException $e) {
      $linkedTo = [];
      foreach($e->getPayload() as $entityName => $items) {
        foreach($items as $item) {
          $message = $translator->trans('media.list.messages.couldNotBeRemoved', array(
            '%entity%' => $this->getEntityName()
          ), 'admin');
          $linkedTo[$message][] = $item->toHumanReadable();
        }
      }

      $this->addFlash(
        'error', $linkedTo
      );

    }

    $route = $request->get('_route');
    return $this->redirectToRoute($route, array(
      '_locale' => $request->getLocale()
    ));
  }

  public function exportAction(Request $request) {
    if (empty($this->exportConfig)) {
      // OVERRIDE TO SET EXPORT CONFIG
      throw new \Exception('Not implemented');
    }
    if ($this->container->has('bazookas.export.service.ExportService')) {
      $exportService = $this->get('bazookas.export.service.ExportService');
      $response = $exportService->export($request, $this->exportConfig);
    } else {
      # Call to old export-service for mobilefans event stuff that happens in there
      $exportService = $this->get('bazookas.admin.service.exportservice');
      $response = $exportService->export($this->exportConfig, $this->event);
    }
    return $response;
  }

  // --------------------------------------------------
  // PROTECTED UTILITY FUNCTIONS
  /**
   * this function is called before anything is executed
   * override this function to check access
   * $this->config->setAccess("add/edit/remove", false) to hide the action buttons
   * @param string $_action the current page action
   * @param mixed $_id the id (primary key) of the current item, for edit pages, can be null
   * @return boolean TRUE to grant access, FALSE to throw access denied error
   **/
  protected function checkAccess($_action, $_id) {
    // TODO use the permissions defined in the entity for this instead of relying on an overwrite!
    // NOTE only use this for globally denying access to certain actions to check per entity use the entity access checks
    return $this->config->getAccess($_action);
  }

  /**
   * @return ObjectRepository a repository instance linked to the admin list entity class
   **/
  protected function getRepository() {
    return $this
      ->getDoctrine()
      ->getManager()
      ->getRepository($this->config->entity);
  }
  /**
   * @return QueryBuilder a querybuilder instance linked to the admin list entity class
   **/
  protected function getQueryBuilder() {
    $repository = $this
      ->getDoctrine()
      ->getManager()
      ->getRepository($this->config->entity);

    return $repository->createQueryBuilder('e');
  }

  /**
   * shared function for handling add/edit form
   * @param Request $request the symfony request object
   * @param Form $form a symfony form instance
   * @param mixed $entity  an instance of the entity to be added/edited
   * @param string $view path to view to be used to render page (add/edit)
   * @param string $pageTitle translation key for the page title
   * @param string $successMessage translation key of the success message to display when adding/editing was a success
   * @param array $returnRouteParams
   * @return Response a response object to return from the controller action
   **/
  protected function handleForm(Request $request, $form, $entity, $view, $pageTitle, $successMessage, $returnRouteParams = []) {
    $form = $this->processForm($request, $form, $entity, $successMessage, $returnRouteParams);

    if ($form instanceof RedirectResponse) {
      return $form;
    }

    $templateData = [
      'form' => $form->createView(),
      'titleTranslationKey' => $pageTitle,
      'titleTranslationVariables' => array('%entity%' => $this->getEntityName()),
    ];
    $templateData = $this->mergeTemplateData($templateData, $request->get('_action'));

    return $this->render($view, $templateData);
  }

  /**
   * This function handles the actual processing of the form and the persisting of the form items
   * @param Request $request
   * @param $form
   * @param $entity
   * @param $successMessage
   * @param array $returnRouteParams
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   */
  protected function processForm(Request $request, $form, $entity, $successMessage, $returnRouteParams = []) {
    $form->handleRequest($request);

    if ($form->isSubmitted() && $form->isValid()) {
      $entityManager = $this->getDoctrine()->getManager();

      $this->event->entity = $entity;
      $this->dispatcher->dispatch(AdminListEvent::BEFORE_PERSIST, $this->event);

      $this->invokePrePersistFunction($entity);
      $entityManager->persist($entity);

      $translator = $this->get('translator');
      try {
        $entityManager->flush();

        $this->event->id = $entity->getId();
        $this->dispatcher->dispatch(AdminListEvent::AFTER_PERSIST, $this->event);

        $route = $request->get('_route');
        // set flash message
        $this->addFlash(
          'success',
          $translator->trans($successMessage, array(
            '%entity%' => $this->getEntityName()
          ), 'admin')
        );

        $routeParams = array(
          '_locale' => $request->getLocale()
        );

        if (!empty($returnRouteParams)) {
          $routeParams = array_merge($routeParams, $returnRouteParams);
        }

        return $this->redirectToRoute($route, $routeParams);

      } catch(AdminExceptionInterface $e) {
        $this->addFlash(
          'error',
          $translator->trans($e->getMessage(), [], 'admin')
        );
      }
    }

    return $form;
  }

  /**
   * Remove a single entity by id, creates a NotFound flash bag message when no entity is found for this id
   * @param mixed $id the id (primary key) of the entity you want to remove
   **/
  protected function removeEntity($id) {
    // first get ourselves an entity manager and the entity
    $em = $this->getDoctrine()->getManager();
    $entity = $em->getRepository($this->config->entity)->find($id);

    // dispatch event
    $this->event->entity = $entity;
    $this->dispatcher->dispatch(AdminListEvent::BEFORE_REMOVE_ENTITY, $this->event);

    // check if the entity exists and throw error if not
    if (!$entity) {
      $translator = $this->get('translator');
      throw $this->createNotFoundException(
        $translator->trans('admin.error.entityNotFound', array(
          '%entity%' => $this->getEntityName(),
          '%id%' => $id
        ), 'admin')
      );
    }

    $em->remove($entity);
    $em->flush();

    $this->dispatcher->dispatch(AdminListEvent::AFTER_REMOVE_ENTITY, $this->event);
  }

  /**
   * Get the translated entity name (using the translation key from config)
   * @return string translated name or key if no translation is found
   **/
  protected function getEntityName() {
    return $this->get('translator')->trans($this->config->entityName, [], 'admin');
  }

  /**
   * Get the translated entity name (using the translation key from config)
   * @return string translated name or key if no translation is found
   **/
  protected function getEntityPluralName() {
    return $this->get('translator')->trans($this->config->entityName.'Plural', [], 'admin');
  }

  // --------------------------------------------------
  // PAGE AND "SINGLE ENTITY" ACTIONS
  /**
   * Create the action buttons for the list page
   * @param Request $request
   **/
  protected function createListActions(Request $request) {
    $route = $request->get('_route');

    if ($this->config->getAccess(self::ACTION_EXPORT)) {
      $this->config->pageActions[] = array(
        'path' => $this->generateUrl($route, array(
          '_action' => self::ACTION_EXPORT
        )),
        'color' => 'warning',
        'icon' => 'ti-download',
        'label' => 'admin.actions.export',
        'action' => self::ACTION_EXPORT
      );
    }

    if ($this->config->getAccess(self::ACTION_VIEW)) {
      $this->config->addAction('BazookasAdminBundle:List\Partials\ListActions:_viewButton.html.twig', $route);
    }

    if ($this->config->getAccess(self::ACTION_EDIT)) {
      $this->config->addAction('BazookasAdminBundle:List\Partials\ListActions:_editButton.html.twig', $route);
    }

    if ($this->config->getAccess(self::ACTION_REMOVE)) {
      $this->config->addAction('BazookasAdminBundle:List\Partials\ListActions:_removeButton.html.twig', $route);
    }

    if ($this->config->getAccess(self::ACTION_ADD)) {
      $path = $this->generateUrl($route, array(
        '_action' => self::ACTION_ADD
      ));

      $this->config->pageActions[] = new PageAction([
        'path' => $path,
        'label' => 'admin.actions.add',
        'color' => 'success',
        'icon' => 'ti-plus'
      ]);
    }

  }

  /**
   * Create the action buttons for the edit page
   * @param Request $request
   **/
  protected function createEditActions(Request $request, $_id) {
    $route = $request->get('_route');
    $attributes = $request->get('_route_params');
    unset($attributes['_action']);
    unset($attributes['_id']);

    $this->config->pageActions[] = new PageAction(
      [
        'path' => $this->generateUrl($route, $attributes),
        'label' => 'admin.actions.cancel',
        'color' => 'warning',
        'icon' => 'ti-arrow-left',
      ]
    );

    $this->config->pageActions[] = new PageAction(
      [
        'path' => '#',
        'color' => 'success',
        'icon' => 'ti-save',
        'label' => 'admin.form.submit',
        'class' => 'js-form-submit',
      ]
    );
  }

  /**
   * Create the action buttons for the edit page
   * @param Request $request
   * @param int $_id
   **/
  protected function createViewActions(Request $request, $_id) {
    $route = $request->get('_route');
    $attributes = $request->get('_route_params');
    unset($attributes['_action']);
    unset($attributes['_id']);

    $this->config->pageActions[] = new PageAction(
      [
        'path' => $this->generateUrl($route, $attributes),
        'color' => 'warning',
        'icon' => 'ti-arrow-left',
        'label' => 'admin.actions.cancel',
      ]
    );
  }

  /**
   * Create the action buttons for the add page
   * @param Request $request
   **/
  protected function createAddActions(Request $request) {
    $route = $request->get('_route');
    $attributes = $request->get('_route_params');
    unset($attributes['_action']);
    unset($attributes['_id']);

    $this->config->pageActions[] = new PageAction(
      [
        'path' => $this->generateUrl($route, $attributes),
        'color' => 'warning',
        'icon' => 'ti-arrow-left',
        'label' => 'admin.actions.cancel',
      ]
    );

    $this->config->pageActions[] = new PageAction(
      [
        'path' => '#',
        'color' => 'success',
        'icon' => 'ti-save',
        'label' => 'admin.form.submit',
        'class' => 'js-form-submit'
      ]
    );
  }

  protected function invokePrePersistFunction($entity) {
    if ($entity instanceof EntityInterface || method_exists($entity, 'preFormPersist')) {
      $entity->preFormPersist();
    }
  }

  protected function invokePreUpdateFunction($entity) {
    if ($entity instanceof EntityInterface || method_exists($entity, 'preFormUpdate')) {
      $entity->preFormUpdate();
    }
  }

  protected function accessDeniedRedirect(Request $request) {
    $translator = $this->get('translator');
    $this->addFlash(
      'error',
      $translator->trans('admin.list.messages.accessDenied', [], 'admin')
    );

    //send the user back where they belong!
    $route = $request->get('_route');

    if ($this->action === self::ACTION_LIST) {
      $route = 'bazookas_admin_homepage';
    }
    else if (empty($route)) {
      throw new AccessDeniedException();
    }

    return $this->redirectToRoute($route, array(
      '_action' => null,
      '_id' => null,
    ));
  }

  protected function getRefererRoute(Request $request) {
    $referer = $request->headers->get('referer');
    $baseUrl = $request->getSchemeAndHttpHost();

    $lastPath = substr($referer, strpos($referer, $baseUrl) + strlen($baseUrl));
    return $this->get('router')->getMatcher()->match($lastPath);
  }

  public function updateOrderAjaxAction(Request $request) {
    $data = json_decode($request->getContent(), true);

    $ids = [];
    foreach($data as $element) {
      $ids[] = $element['id'];
    }

    $entities = $this->getRepository()->getEntitiesById($ids);

    //Execute access controle check to make sure the user is allowed to perform the action
    $auth = $this->get('security.authorization_checker');

    foreach($entities as $key => &$entity) {
      if (empty($entity) || !$entity->canEdit($auth, $this->getUser())) {

        //TODO handle this in a better way?
        $response = new JsonResponse();
        $response->setStatusCode(401);
        $response->setData(array(
          'You do not have the necessary rights to modify these entities.'
        ));
        return $response;
      }
    }

    foreach($data as $element) {
      foreach($entities as &$entity) {
        if ($entity->getId() === $element['id']) {
          $entity->setOrder($element['order']);
          break;
        }
      }
    }

    $this->getRepository()->updateEntities($entities);

    $response = new JsonResponse();
    return $response;
  }

  /**
   * This function provides a means of adding additional template data to a controller.
   * @param array $templateData
   * @param string $action
   * @return array $templateData
   */
  protected function mergeTemplateData(array $templateData, $action) {
    return array_merge($templateData, [
      'config' => $this->config,
      'pageActions' => $this->config->pageActions,
      'styles' => $this->config->getStyles(),
      'scripts' => $this->config->getScripts(),

      //TODO this is deprecated, to be removed
      'custom' => $this->customViewData,
    ]);
  }

}
