<?php

namespace Bazookas\AdminBundle\Filter;

use Bazookas\AdminBundle\Form\ListFilterCollectionForm;
use Doctrine\ORM\QueryBuilder;
use Symfony\Component\Form\FormFactory;
use Symfony\Component\HttpFoundation\Request;

class ListFilterService
{

  /**
   * @var ListFilter[]
   */
  protected $filters;

  /**
   * @var bool
   */
  protected $maintainFilterState = false;

  public function __construct()
  {
    $this->filters = [];
  }

  public function getFilters()
  {
    return $this->filters;
  }

  public function addFilter($fieldName, ListFilter $filter)
  {
    $this->filters[$fieldName] = $filter;

    return $this;
  }

  protected function hasFilter($fieldName)
  {
    return isset($this->filters[$fieldName]);
  }

  public function getFilter($fieldName)
  {
    if ($this->hasFilter($fieldName)) {
      return $this->filters[$fieldName];
    }

    return null;
  }

  public function getMaintainFilterState()
  {
    return $this->maintainFilterState;
  }

  public function setMaintainFilterState($maintainFilterState)
  {
    $this->maintainFilterState = $maintainFilterState;

    return $this;
  }

  /**
   * This function parses the filters from the request object
   * @param Request $request
   *
   * @return $this
   */
  public function parseFilters(Request $request)
  {
    $route = $request->get('_route');
    $action = $request->get('_action');
    $session = $request->getSession();
    $filterHash = md5($route . $action);

    $clear = $request->query->get('clearFilters');

    if (!isset($clear)) {
      //handle any other filters passed in the url
      $filters = $request->query->get('filter', []);
      $filterForm = $request->query->get('filter_form', []);

      //clear any existing filters
      $this->filters = [];

      if (!empty($filterForm['filters'])) {
        $filters = $filterForm['filters'];
      }

      foreach ($filters as $filter) {
        $filterObject = new ListFilter();
        $filterObject->setField($filter['field']);
        $filterObject->setOperator($filter['operator']);
        if (isset($filter['value'])) {
          $filterObject->setValue($filter['value']);
        } else {
          $filterObject->setValue(0);
        }

        if (isset($filter['type'])) {
          $filterObject->setType($filter['type']);
        }

        $this->filters[] = $filterObject;
      }

      if (empty($this->filters) && $this->getMaintainFilterState()) {
        $sessionFilters = $session->get('filters');

        if (isset($sessionFilters[$filterHash])) {
          $this->filters = $sessionFilters[$filterHash];
          return $this;
        } else if (!empty($sessionFilters)) {
          $session->remove('filters');
        }
      }

      //overwrite the session storage, this will remove any other filters still present
      $session->set('filters', [$filterHash => $this->filters]);
    } else {
      $this->filters = [];
      $session->set('filters', [$filterHash => $this->filters]);
    }

    return $this;
  }

  /**
   * This function executes the filters
   * @param queryBuilder $queryBuilder the querybuilder instance to apply filters to
   *
   * @return $this
   */
  public function executeFilters(QueryBuilder &$queryBuilder)
  {
    //Example of filter url
    //http://localhost:8001/en/admin/lessonplan/list?filter[0][field]=details.title&filter[0][operator]=like&filter[0][value]=test

    $filterParameter = 1;
    foreach ($this->filters as &$filter) {

      if (is_object($filter)) {
        $field = $this->joinAndGetFieldName($queryBuilder, $filter->getField());
        $operator = $filter->getOperator();
        $value = $filter->getValue();
        $type = $filter->getType();
      } else {
        $field = $this->joinAndGetFieldName($queryBuilder, $filter['field']);
        $operator = $filter['operator'];
        $value = $filter['value'];
        $type = isset($filter['type']) ? $filter['type'] : ListFilterField::TYPE_TEXT;
      }

      if ($value === null) {
        $operator = 'isNull';
      }
      else if ($type === ListFilterField::TYPE_DATETIME) {
        $value = new \DateTime($value, new \DateTimeZone('Europe/Brussels'));
      }

      if ($operator === 'like' || $operator === 'notLike') {
        $value = "%$value%";
      }

      $queryBuilder->andWhere(
        $queryBuilder->expr()->{$operator}($field, ":filter$filterParameter")
      );
      if ($operator != 'isNull') {
        $queryBuilder->setParameter(":filter$filterParameter", $value);
      }

      $filterParameter++;
    }

    return $this;
  }

  private function joinAndGetFieldName(QueryBuilder $queryBuilder, $fieldPath)
  {
    $exploded = explode('.', $fieldPath);
    if (count($exploded) === 1) {
      return 'e.' . $fieldPath;
    } else {
      $lastField = array_pop($exploded);
      $previousField = 'e';
      foreach ($exploded as $field) {
        if (!$this->joinExsists($queryBuilder, $previousField, $field)) {
          // add the join to the queryBuilder
          $queryBuilder->leftJoin("$previousField.$field", $field);

          // add the joined entity to the select statement
          // to make sure the joined entities stay filtered
          $queryBuilder->addSelect($field);
        }
        $previousField = $field;
      }

      return "$field.$lastField";
    }
  }

  private function joinExsists(QueryBuilder $queryBuilder, $prefix, $alias)
  {
    $parts = $queryBuilder->getDQLPart('join');
    if (isset($parts[$prefix])) {
      foreach ($parts[$prefix] as $existingJoin) {
        if ($existingJoin->getAlias() === $alias) {
          return true;
        }
      }
    }

    return false;
  }

  public function createFilterForm(FormFactory $formFactory, array $filterFields) {
    if (!empty($filterFields)) {
      $filters = $this->getFilters();

      if (empty($filters)) {
        $listFilter = new ListFilter();
        $listFilter
          ->setFilterFields($filterFields)
        ;
        $filters = [$listFilter];
      }
      else {
        foreach ($filters as $filter) {
          $filter->setFilterFields($filterFields);
        }
      }

      $filterCollection = new ListFilterCollection();
      $filterCollection->setFilters($filters);

      return $formFactory
        ->createNamedBuilder('filter_form', ListFilterCollectionForm::class, $filterCollection, array(
          'method' => 'GET',
        ))
        ->getForm()
      ;
    }

    return null;
  }
}
