<?php

namespace Bazookas\AdminBundle\PageBuilder;

use Bazookas\AdminBundle\AdminElements\Elements\FilterFormPanelElement;
use Bazookas\AdminBundle\AdminElements\Elements\HeaderElement;
use Bazookas\AdminBundle\AdminElements\Elements\ListFieldElement;
use Bazookas\AdminBundle\AdminElements\Containers\RowContainer;
use Bazookas\AdminBundle\AdminElements\Interfaces\AdminElementInterface;
use Bazookas\AdminBundle\Controller\Base\BaseAdminActionController;
use Bazookas\AdminBundle\Filter\ListFilter;
use Bazookas\AdminBundle\Filter\ListFilterField;
use Bazookas\AdminBundle\Filter\ListFilterService;
use Bazookas\AdminBundle\AdminElements\Containers\ListTableContainer;
use Bazookas\AdminBundle\PageBuilder\Base\BasePageBuilder;
use Bazookas\AdminBundle\PageBuilder\Interfaces\ListPageBuilderInterface;
use Bazookas\AdminBundle\Repository\Interfaces\ListRepositoryInterface;
use Bazookas\CommonBundle\Exception\NotImplementedException;
use Doctrine\ORM\Query;
use Doctrine\ORM\QueryBuilder;
use Symfony\Component\DependencyInjection\ContainerAwareInterface;
use Symfony\Component\DependencyInjection\ContainerAwareTrait;
use Symfony\Component\Form\Form;
use Symfony\Component\HttpFoundation\Request;

class ListPageBuilder extends BasePageBuilder implements ContainerAwareInterface, ListPageBuilderInterface
{
  use ContainerAwareTrait;

  /**
   * @var ListFieldElement[]
   */
  protected $fields = [];

  /**
   * @var array
   */
  protected $listQueryParams = [];

  /**
   * @var integer
   */
  protected $limit = 20;

  /**
   * @var ListFilterField[]
   */
  protected $filterFields = [];

  /**
   * @var bool
   */
  protected $maintainFilterState = false;

  /**
   * @var ListFilterService
   */
  protected $filterService;

  /**
   * @var bool
   */
  private $ajaxPageable = false;

  /**
   * @var string
   */
  private $pageAction = BaseAdminActionController::ACTION_LIST;

  /**
   * @var bool
   */
  private $inlineEditable;

  /**
   * @return ListFieldElement[]
   */
  public function getFields()
  {
    return $this->fields;
  }

  /**
   * @param ListFieldElement[] $fields
   * @return $this
   */
  public function setFields(array $fields)
  {
    $this->fields = $fields;

    return $this;
  }

  /**
   * @param string $fieldName
   * @param array $options
   * @return $this
   */
  public function addField($fieldName, array $options = []) {
    $options['field'] = $fieldName;

    if (!isset($options['label'])) {
      $options['label'] = 'admin.entities.' . lcfirst($this->getEntityShortClassName()) . '.fields.' . $fieldName;
    }

    if (!isset($options['inlineEditable'])) {
      $options['inlineEditable'] = $this->inlineEditable;
    }

    $this->fields[] = new ListFieldElement([], $options);
    return $this;
  }

  /**
   * @param string $fieldName
   * @param array $options
   * @return $this
   */
  public function addBooleanField($fieldName, array $options = []) {
    if (!isset($options['template'])) {
      $options['template'] = 'BazookasAdminBundle:List/Fields:_boolean.html.twig';
    }

    $this->addField($fieldName, $options);
    return $this;
  }

  /**
   * @param string $fieldName
   * @param array $options
   * @return $this
   */
  public function addDateField($fieldName, array $options = []) {
    if (!isset($options['template'])) {
      $options['template'] = 'BazookasAdminBundle:List/Fields:_date.html.twig';
    }

    $this->addField($fieldName, $options);
    return $this;
  }

  /**
   * @param string $fieldName
   * @param array $options
   * @return $this
   */
  public function addDateTimeField($fieldName, array $options = []) {
    if (!isset($options['template'])) {
      $options['template'] = 'BazookasAdminBundle:List/Fields:_dateTime.html.twig';
    }

    $this->addField($fieldName, $options);
    return $this;
  }

  /**
   * @param string $fieldName
   * @param array $options
   * @return $this
   */
  public function addImageField($fieldName, array $options = []) {
    if (!isset($options['template'])) {
      $options['template'] = 'BazookasAdminBundle:List/Fields:_image.html.twig';
    }

    $this->addField($fieldName, $options);
    return $this;
  }

  /**
   * @param string $fieldName
   * @param array $options
   * @return $this
   */
  public function addImageIconField($fieldName, array $options =[]) {
    if (!isset($options['template'])) {
      $options['template'] = 'BazookasAdminBundle:List/Fields:_imageIcon.html.twig';
    }

    $this->addField($fieldName, $options);
    return $this;
  }

  /**
   * @param string $fieldName
   * @param array $options
   * @return $this
   */
  public function addSortableField($fieldName, array $options = []) {
    if (!isset($options['sortable'])) {
      $options['sortable'] = true;
    }

    $this->addField($fieldName, $options);
    return $this;
  }

  /**
   * @param string $fieldName
   * @param array $options
   * @return $this
   */
  public function addOrderableField($fieldName, array $options = []) {
    if (!isset($options['orderable'])) {
      $options['orderable'] = true;
    }

    $options['field'] = $fieldName;

    if (empty($options)) {
      $options['label'] = 'admin.entities.' . lcfirst($this->getEntityShortClassName()) . '.fields.' . $fieldName;
    }

    $field = new ListFieldElement([], $options);
    $this->fields[] = $field;

    //Make sure the orderable field is placed first
    array_unshift($this->fields, $field);

    return $this;
  }

  /**
   * @return ListFieldElement[]
   */
  protected function getListFields() {
    $listFields = [];
    foreach($this->fields as $field) {
      $listFields[] = $field->getField();
    }

    return $listFields;
  }

  /**
   * @return int
   */
  public function getLimit()
  {
    return $this->limit;
  }

  /**
   * @param int $limit
   * @return $this
   */
  public function setLimit($limit)
  {
    $this->limit = $limit;

    return $this;
  }

  /**
   * @return array
   */
  public function getListQueryParams()
  {
    return $this->listQueryParams;
  }

  /**
   * @param array $listQueryParams
   * @return $this
   */
  public function setListQueryParams(array $listQueryParams)
  {
    $this->listQueryParams = $listQueryParams;

    return $this;
  }

  /**
   * @return ListFilterField[]
   */
  public function getFilterFields()
  {
    return $this->filterFields;
  }

  /**
   * @param ListFilterField[] $filterFields
   * @return $this
   */
  public function setFilterFields(array $filterFields)
  {
    $this->filterFields = $filterFields;

    return $this;
  }

  /**
   * @param string $fieldName
   * @param array $options
   * @return $this
   */
  public function addFilterField($fieldName, array $options = []) {
    $options['field'] = $fieldName;

    if (!isset($options['label'])) {
      $options['label'] = 'admin.entities.' . lcfirst($this->getEntityShortClassName()) . '.fields.' . $fieldName;
    }

    $this->filterFields[] = new ListFilterField($options);
    return $this;
  }

  /**
   * @param string $fieldName
   * @param array $options
   * @return ListPageBuilder
   */
  public function addBooleanFilterField($fieldName, array $options = []) {
    $options['type'] = ListFilterField::TYPE_BOOLEAN;

    if (!isset($options['operators'])) {
      $options['operators'] = [
        'admin.filter.operator.equals' => ListFilter::OPERATOR_EQUALS
      ];
    }

    return $this->addFilterField($fieldName, $options);
  }

  /**
   * @param string $fieldName
   * @param array $options
   * @return ListPageBuilder
   */
  public function addTextFilterField($fieldName, array $options = []) {
    $options['type'] = ListFilterField::TYPE_TEXT;

    if (!isset($options['operators'])) {
      $options['operators'] = [
        'admin.filter.operator.like' => ListFilter::OPERATOR_LIKE,
        'admin.filter.operator.notLike' => ListFilter::OPERATOR_NOT_LIKE,
        'admin.filter.operator.equals' => ListFilter::OPERATOR_EQUALS,
        'admin.filter.operator.notEquals' => ListFilter::OPERATOR_NOT_EQUALS
      ];
    }

    return $this->addFilterField($fieldName, $options);
  }

  /**
   * @param string $fieldName
   * @param array $options
   * @return ListPageBuilder
   */
  public function addDateTimeFilterField($fieldName, array $options = []) {
    $options['type'] = ListFilterField::TYPE_DATETIME;

    if (!isset($options['operators'])) {
      $options['operators'] = [
        'admin.filter.operator.like' => ListFilter::OPERATOR_LIKE,
        'admin.filter.operator.notLike' => ListFilter::OPERATOR_NOT_LIKE,
        'admin.filter.operator.equals' => ListFilter::OPERATOR_EQUALS,
        'admin.filter.operator.notEquals' => ListFilter::OPERATOR_NOT_EQUALS,
        'admin.filter.operator.greaterThan' => ListFilter::OPERATOR_GREATER_THAN,
        'admin.filter.operator.greaterThanEquals' => ListFilter::OPERATOR_GREATER_THAN_EQUALS,
        'admin.filter.operator.lessThan' => ListFilter::OPERATOR_LESS_THAN,
        'admin.filter.operator.lessThanEquals' => ListFilter::OPERATOR_LESS_THAN_EQUALS
      ];
    }

    return $this->addFilterField($fieldName, $options);
  }

  /**
   * @param string $fieldName
   * @param array $options
   * @return ListPageBuilder
   */
  public function addChoiceFilterField($fieldName, array $options = []) {
    $options['type'] = ListFilterField::TYPE_CHOICE;

    if (!isset($options['operators'])) {
      $options['operators'] = [
        'admin.filter.operator.equals' => ListFilter::OPERATOR_EQUALS,
        'admin.filter.operator.notEquals' => ListFilter::OPERATOR_NOT_EQUALS
      ];
    }

    return $this->addFilterField($fieldName, $options);
  }

  /**
   * @param string $fieldName
   * @param array $options
   * @return ListPageBuilder
   */
  public function addNumberFilterField($fieldName, array $options = []) {
    $options['type'] = ListFilterField::TYPE_NUMBER;

    if (!isset($options['operators'])) {
      $options['operators'] = [
        'admin.filter.operator.equals' => ListFilter::OPERATOR_EQUALS,
        'admin.filter.operator.notEquals' => ListFilter::OPERATOR_NOT_EQUALS,
        'admin.filter.operator.greaterThan' => ListFilter::OPERATOR_GREATER_THAN,
        'admin.filter.operator.greaterThanEquals' => ListFilter::OPERATOR_GREATER_THAN_EQUALS,
        'admin.filter.operator.lessThan' => ListFilter::OPERATOR_LESS_THAN,
        'admin.filter.operator.lessThanEquals' => ListFilter::OPERATOR_LESS_THAN_EQUALS
      ];
    }

    return $this->addFilterField($fieldName, $options);
  }

  /**
   * @return bool
   */
  public function getMaintainFilterState()
  {
    return $this->maintainFilterState;
  }

  /**
   * @param bool $maintainFilterState
   * @return $this
   */
  public function setMaintainFilterState($maintainFilterState)
  {
    $this->maintainFilterState = $maintainFilterState;

    return $this;
  }

  /**
   * @return bool
   */
  public function isAjaxPageable()
  {
    return $this->ajaxPageable;
  }

  /**
   * @param bool $ajaxPageable
   * @param null $pageAction
   * @return $this
   */
  public function setAjaxPageable($ajaxPageable, $pageAction = null)
  {
    $this->ajaxPageable = $ajaxPageable;

    if (!empty($pageAction)) {
      $this->pageAction = $pageAction;
    }

    return $this;
  }

  /**
   * @return string
   */
  public function getPageAction()
  {
    return $this->pageAction;
  }

  /**
   * @param string $pageAction
   * @return $this
   */
  public function setPageAction($pageAction)
  {
    $this->pageAction = $pageAction;

    return $this;
  }

  /**
   * @return bool
   */
  public function isInlineEditable()
  {
    return $this->inlineEditable;
  }

  /**
   * @param bool $inlineEditable
   * @return $this
   */
  public function setInlineEditable($inlineEditable)
  {
    $this->inlineEditable = $inlineEditable;

    return $this;
  }

  /**
   * @return ListFilterService
   */
  public function getFilterService()
  {
    if (empty($this->filterService)) {
      $this->filterService = new ListFilterService();
    }

    return $this->filterService;
  }

  /**
   * @param ListFilterService $filterService
   * @return $this
   */
  public function setFilterService(ListFilterService $filterService)
  {
    $this->filterService = $filterService;

    return $this;
  }

  /**
   * @inheritDoc
   */
  public function buildPage(Request $request) {
    //Add a header element
    $this->elements[] = $this->generateHeaderElement();

    //Add the list table
    $this->elements[] = $this->buildListTable($request);
  }

  /**
   * @inheritDoc
   */
  public function buildListTable(Request $request) {
    $listTable = new ListTableContainer([], [
      'ajaxPageable' => $this->isAjaxPageable(),
      'pageAction' => $this->getPageAction(),
      'inlineEditable' => $this->isInlineEditable()
    ]);
    $listTable->setFields($this->fields);
    $listTable->setActions($this->actions);

    $entityRepository = $this->entityManager->getRepository($this->entityClass);

    if (!$entityRepository instanceof ListRepositoryInterface) {
      throw new \Exception('The entity repository for \''. $this->entityClass .'\' must implement the interface ' . ListRepositoryInterface::class  );
    }

    //Get the fields in the table
    $listFields = $this->getListFields();

    //Begin creating the list table query
    $qb = $entityRepository->createListQueryBuilder($listFields, $this->getListQueryParams());

    //apply filtering as needed
    $this->handleFiltering($request, $qb);

    //TODO add sorting stuff


    //Add a before list query hook
    $qb = $entityRepository->beforeListQueryExecution($qb, $this->getListQueryParams());

    //Get the current page
    $page = $request->get('page', 1);
    $page--;

    //Get the list table result
    $data = $entityRepository->getPageableResult($qb, $page, $this->limit, Query::HYDRATE_OBJECT);

    //Fill the list table container with data
    $listTable
      ->setTableData($data['results'])
      ->setPageCount($data['pageCount'])
      ->setCurrentPage($page + 1)
    ;

    return $listTable;
  }

  /**
   * Generate the header element
   * @return AdminElementInterface
   */
  protected function generateHeaderElement() {
    return new RowContainer([
      new HeaderElement([], [
        'label' => 'admin.entities.' . lcfirst($this->getEntityShortClassName()) . '.listLabel'
      ]),
      $this->createPageActionContainer()
    ], [
      'defaultColClass' => 'col-md-6'
    ]);
  }

  /**
   * @param Request $request
   * @param QueryBuilder $qb
   */
  protected function handleFiltering(Request $request, QueryBuilder $qb = null) {
    $filterService = $this->getFilterService();
    $filterService
      ->setMaintainFilterState($this->getMaintainFilterState())
      ->parseFilters($request)
      ->executeFilters($qb, $this->filterFields)
    ;

    //create and add the filter form
    $this->addFilterForm($filterService);
  }

  /**
   * @param ListFilterService $filterService
   */
  protected function addFilterForm(ListFilterService $filterService) {
    //Create a filter form
    $formFactory = $this->container->get('form.factory');
    $filterFields = $this->getFilterFields();

    if (!empty($filterFields)) {
      $filterForm = $filterService->createFilterForm($formFactory, $this->getFilterFields());
      if (!empty($filterForm) && $filterForm instanceof Form) {
        $this->elements[] = new FilterFormPanelElement([], [
          'form' => $filterForm->createView(),
          'expanded' => !empty($filterService->getFilters()),
        ]);
      }
    }
  }
}
