## Creating and using a custom CMS user

This guide will attempt to explain how to override the default CMSUser so that you can use your own user instead. Note that it is **not** possible to use the default CMSUser in conjunction with your own CMSUser.

First start by creating your own user:

```php
<?php

namespace WebsiteBundle\Entity;

use Bazookas\AdminBundle\Entity\CMSUser;
use Doctrine\Common\Collections\ArrayCollection;
use Doctrine\ORM\Mapping as ORM;
use Symfony\Component\Validator\Constraints as Assert;

/**
 * @ORM\Entity(repositoryClass="WebsiteBundle\Repository\ResearcherRepository")
 * @ORM\HasLifecycleCallbacks
 */
class Researcher extends CMSUser {
  //Project specific stuff
}
```

The important things to note here are that your CMSUser must extend the `CMSUser` class and must define its own Repository class. The repository class must implement `Bazookas\AdminBundle\Repository\Interfaces\ListRepositoryInterface`. You can create your service and extend the admin bundle's `BaseRepository` to make use of a default implementation of the interface.

```php
<?php

namespace WebsiteBundle\Repository;

use Bazookas\AdminBundle\Repository\Base\BaseRepository;
use WebsiteBundle\Entity\Researcher;

class ResearcherRepository extends BaseRepository
{

  public function getEntityClass() {
    return Researcher::class;
  }
  
}
```

Once you have your entity and its related repository; start by creating a new list controller:

```php
<?php

namespace WebsiteBundle\Controller\Admin;

use Bazookas\AdminBundle\Controller\AdminController\BaseCMSUserAdminController;
use Bazookas\AdminBundle\Controller\Base\BaseAdminListController;
use Bazookas\AdminBundle\PageBuilder\ListPageBuilder;
use Symfony\Component\HttpFoundation\Request;
use WebsiteBundle\Entity\Researcher;
use WebsiteBundle\Form\Admin\ResearcherAdminType;

class ResearcherAdminController extends BaseCMSUserAdminController
{

  public function getEntityClass()
  {
    return Researcher::class;
  }

  protected function modifyListBuilder(Request $request, ListPageBuilder $builder) {
    $builder = BaseAdminListController::modifyListBuilder($request, $builder);

    $builder
      ->addField('username')
      ->addField('email')
      ->addField('groups', [
        'template' => 'BazookasAdminBundle:List/Fields/CMSUser:_groups.html.twig'
      ])
      
      //add your project specific fields here
    ;

    return $builder;
  }

  protected function getFormClass($action) {
    switch($action) {
      case self::ACTION_ADD:
        return ResearcherAdminType::class;
      case self::ACTION_EDIT:
        return ResearcherAdminType::class;
      default:
        return null;
    }
  }
}
```

As you know any list controller should extend the `BaseAdminListController` class. However a CMSUser specific base class also exists which in most cases you should extend directly (`BaseCMSUserAdminController`). As you can see in the above example it is now possible to add new fields, remove default fields or define custom add or edit classes. In essence you can now perform the exact same actions as a standard `BaseAdminListController` is capable of.

Now as a final step to get your custom CMSUser to work you must configure it be used by the FOSUserBUndle. This allows your new class to work with the cms's security system. In `config.yml` change the following setting to:

```yaml
fos_user:
    user_class: WebsiteBundle\Entity\Researcher
```

Congratulations! The custom user should now be working! :)
