## Elements and containers

Each element or container can always be given the following options:
```php
  protected function getDefaultOptions() {
    return [
      'roles' => ['ROLE_ADMIN'],
      'template' => '',
      'class' => '',
      'value' => null,
      'label' => '',
      'colClass' => '',
      'icon' => '', // for things like tabbed panel or flippable container,
      'iconClass' => '',
      'navName' => ''
    ];
  }
```
Any additional options which can be used with specific elements is mentioned in their documentation below.

Elements: items that will be rendered directly

- [AlertElement](#alertelement)
- [CalendarElement](#calendarelement)
- [ErrorElement](#errorelement)
- [FormElement](#formelement)
- [GraphElement](#graphelement)
- [IconTextElement](#icontextelement)
- [ImageElement](#imageelement)
- [ProfileElement](#profileelement)
- [ProgressedStatisticElement](#progressedstatisticelement)
- [StatisticElement](#statisticelement)
- [StatisticOverviewElement](#statisticoverviewelement)
- [StatisticPanelElement](#statisticPanelElement)
- [TextElement](#textelement)
- [UserStatisticElement](#userstatisticelement)

Containers: items that are meant to hold other elements

- [Container](#container)
- [ListContainer](#listcontainer)
- [PanelContainer](#panelcontainer)
- [RowContainer](#rowcontainer)
- [TabbedContainer](#tabbedcontainer)
- [TabbedPanelContainer](#tabbedpanelcontainer)
- [TableContainer](#tablecontainer)
- [WellContainer](#wellcontainer)
- [FlippablePanelContainer](#flippablepanelcontainer)

### Elements

#### AlertElement

An alert element is an element which adds a bootstrap alert to the page.

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard\Elements:alertElement.html.twig',  //the template to be used by the widget
      'class' => '',  //a class on the parent element of the widget
      'message' => '',  //The message to be shown
      'type' => self::TYPE_INFO,  //The type of message
      'isDismissible' => true //Is the alert dismissible?
    ]);
  }
```

Available types:
```php
  const TYPE_INFO = 'info';
  const TYPE_DANGER = 'danger';
  const TYPE_SUCCESS = 'success';
  const TYPE_PRIMARY = 'primary';
```

Example:
```php
  $element = new AlertElement([], [
    'message' => 'admin.errors.miniGameNotFound',
    'type' => AlertElement::TYPE_DANGER,
  ]);
```

#### CalendarElement

The calendar element renders a basic 12 month calendar (current month + 11). It is currently not possible to change the range of the months to be rendered. 

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard\Elements:calendarElement.html.twig',
      'class' => 'panel panel-colorful',
      'panelColor' => 'primary',
      'icon' => 'fa-user',
      'iconClass' => 'fa fa-4x mar-all',
      'data' => []
    ]);
  }
``` 

The calendar expects the data to be in the format:
```php
[
    'date' => new \DateTime(),
    'label' => '',
    'colorClass' => 'success',
    'link' => ''
]
```

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard\Elements:errorElement.html.twig',
      'class' => 'panel panel-colorful',
      'panelColor' => 'danger',
      'icon' => 'fa fa-exclamation-triangle',
      'iconClass' => 'fa-4x mar-all'
    ]);
  }
```

Example:
```php

$data = [
    [
        'date' => new \DateTime(),
        'label' => 'an event',
        'colorClass' => 'success',
        'link' => ''
    ]
]

$calendar = new CalendarElement([], ['data' => $data]);
```

#### ErrorElement

This element is used to render an error widget.

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard\Elements:errorElement.html.twig',
      'class' => 'panel panel-colorful',
      'panelColor' => 'danger',
      'icon' => 'fa fa-exclamation-triangle',
      'iconClass' => 'fa-4x mar-all'
    ]);
  }
```

Example:
```php
$element = new ErrorElement([], [
  'label' => 'dashboard.main.tip.noTipError',
]);
```

#### FormElement
This element is used to render a form in a panel.

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard\Elements:formElement.html.twig',
      'class' => '',
      'form' => null
    ]);
  }
```

Example:
```php
$element = new FormElement([], [
  'form' => $mySymfonyForm,
]);
```

#### GraphElement

An element to render a graph item.

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard/Elements:graphElement.html.twig',
      'class' => 'canvas-responsive js-graph',
      'type' => self::TYPE_BAR,
      'options' => [],
      'data' => [],
      'icon' => 'fa-pie',
      'iconClass' => 'fa'
    ]);
  }
```

available graph types:

```php
  const TYPE_RADAR = 'radar';
  const TYPE_BAR = 'bar';
  const TYPE_HORIZONTAL_BAR = 'horizontalBar';
  const TYPE_PIE = 'pie';
  const TYPE_DOUGHNUT = 'doughnut';
```

Example:

```php
$graphTitle = $this->translator->trans('admin.entities.appUser.views.stats.miniGameGraphTitle');

$dataDefinition = GraphElement::generateDefaultDataDefinition(GraphElement::TYPE_RADAR);
$dataDefinition['datasets'][0]['data'] = $data;
$dataDefinition['datasets'][0]['label'] = $graphTitle;
$dataDefinition['labels'] = $labels;

return new RowContainer(
  new GraphElement([], [
    'type' => GraphElement::TYPE_PIE,
    'data' => $dataDefinition,
    'options' => [
      'title' => [
        'display' => true,
        'text' => $graphTitle
      ],
      'tooltips' => [
        'callbacks' => [
          'label' => 'percentageTooltip'
        ]
      ],
      'animation' => [
       'onComplete' => 'addPercentageToSegment',
       'onProgress' => 'addPercentageToSegment',
     ],
     'hover' => [
       'animationDuration' => 0,
     ],
    ]
  ])
}
```

A graph item can also be given the following callback functions:

`percentageTooltip` => add a `%` sign to the end of the value in the tooltip
`addPercentageToSegment` => add the pie charts percent value to the drawn segment.
`addValueToBar` => add the bar graphs value to the drawn segment.

#### IconTextElement

Renders an icon element.

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard\Elements:iconTextElement.html.twig',
      'class' => 'text-lg text-semibold text-main',
      'iconClass' => 'text-lg text-semibold text-main'
    ]);
  }
```

A sample implementation:
```php
$element = new IconTextElement([], [
  'value' => 'the text'
]);
```

#### ImageElement

Renders an image.

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'roles' => ['ROLE_ADMIN'],
      'template' => 'BazookasAdminBundle:Dashboard\Elements:imageElement.html.twig',
      'class' => 'img-md',
    ]);
  }
```

```php
new ImageElement([], [
  'avatarId' => $user->getAvatarId(),
  'class' => 'img-md img-circle mar-btm',
  'template' => 'WebsiteBundle:Dashboard\\AppUser:imageElement.html.twig',
]);
```

#### ProfileElement

Renders a profile element.

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard\Elements:profileElement.html.twig',
      'class' => 'panel',
    ]);
  }
```

Example:
```php
new ProfileElement(
  [
    new ListContainer(
      [
        new TextElement([], [
          'value' => $user->getStoryMode(),
          'label' => 'admin.entities.appUser.fields.storyMode',
          'colClass' => 'col-xs-4 pad-all',
          'template' => 'WebsiteBundle:Dashboard\\AppUser:storyModeElement.html.twig',
        ]),
        new TextElement([], [
          'value' => $user->getCurrentTier() + 1,
          'label' => 'admin.entities.appUser.fields.currentTier',
          'colClass' => 'col-xs-4 pad-all',
        ]),
        new TextElement([], [
          'value' => $user->getRankScore(),
          'label' => 'admin.entities.appUser.fields.rankScore',
          'colClass' => 'col-xs-4 pad-all',
        ]),
      ]
    ),
    new ListContainer(
      [
        new TextElement([], [
          'value' => $user->getRank(),
          'label' => 'admin.entities.appUser.fields.rank',
          'colClass' => 'col-xs-4 pad-all',
        ]),
        new TextElement([], [
          'value' => $user->getGamesPlayed(),
          'label' => 'admin.entities.appUser.fields.gamesPlayed',
          'colClass' => 'col-xs-4 pad-all',
        ]),
      ]
    ),
  ],
  [
    'image' => new ImageElement([], [
      'avatarId' => $user->getAvatarId(),
      'class' => 'img-md img-circle mar-btm',
      'template' => 'WebsiteBundle:Dashboard\\AppUser:imageElement.html.twig',
    ]),
    'name' => $user->getUsername(),
    'role' => null,
  ]
);
```

#### ProgressedStatisticElement
Used to render a statistic widget with a progress bar beneath it.

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard\Elements:progressedStatisticElement.html.twig',
      'class' => 'panel panel-colorful',
      'panelColor' => 'primary',
      'icon' => 'fa-user',
      'iconClass' => 'fa fa-4x',
      'progressValue' => 0,
      'progressLabel' => '',
    ]);
  }
```

Example:
```php
$element = new ProgressedStatisticElement([], [
    'value' => $count,
    'label' => 'dashboard.main.login.activeUsers',
    'icon' => 'fa-user',
    'progressValue' => $percent,
    'progressLabel' => 'dashboard.main.login.progress',
    'panelColor' => 'success'
]);
```

#### StatisticElement
Used to render a statistic widget (For use with a single statistic).

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard\Elements:statisticElement.html.twig',
      'class' => 'panel panel-colorful',
      'panelColor' => 'primary',
      'icon' => 'fa-user',
      'iconClass' => 'fa fa-4x mar-all',
    ]);
  }
```

Example:
```php
$element = new StatisticElement([], [
  'value' => $teamCount,
  'label' => 'dashboard.main.team.count',
  'icon' => 'fa-users',
  'panelColor' => 'primary'
]);

```

#### StatisticOverviewElement
Used to render a widget containing a series of statistics (rendered as a table within a panel). The element will automatically highlight the largest value in green. The lowest value is highlighted in red. All other values are highlighted in blue.

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard\Elements:statisticOverviewElement.html.twig',
      'class' => 'panel panel-colorful',
      'panelColor' => 'primary',
      'icon' => 'fa-user',
      'iconClass' => 'fa fa-4x mar-all',
      'data' => []
    ]);
  }
```

Example:
```php
$data = [
    [
        'label' => 'a label',
        'value' => 3
    ]
];

$element = new StatisticOverviewElement([], [
  'label' => 'dashboard.main.playedCount.label',
  'value' => $total,
  'panelColor' => 'info',
  'icon' => 'fa-gamepad',
  'data' => $data
]);
```

#### StatisticPanelElement


//TODO not implemented


#### TextElement

Renders a text element.

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard\Elements:textElement.html.twig',
      'class' => '',
    ]);
  }
```

Example:
```php
new TextElement([], [
  'value' => $user->getStoryMode(),
  'label' => 'admin.entities.appUser.fields.storyMode',
  'colClass' => 'col-xs-4 pad-all',
  'template' => 'WebsiteBundle:Dashboard\\AppUser:storyModeElement.html.twig',
]);
```

#### UserStatisticElement
Used to render user related statistics.

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard\Elements:userStatisticElement.html.twig',
      'class' => 'panel panel-colorful',
      'panelColor' => 'success'
    ]);
  }
```

Example:
```php
$element = new UserStatisticElement([], [
  'value' => 1000,
  'label' => 'Highscore'
]);
```

### Containers
#### Container

This element has no other purpose then rendering its children and should therefore be used to functionally group items within other containers or elements.

```php
new Container($childElements, $options);
```

#### ListContainer

Renders the children as a list.

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard/Containers:listContainer.html.twig',
      'class' => 'list-unstyled text-center mar-no row'
    ]);
  }
```

If a child is given a `colClass` option then this will be taken into account when rendering the child. The col class is a bootstrap column class.

Example:
```php
new ListContainer(
    [
    new TextElement([], [
      'value' => $user->getStoryMode(),
      'label' => 'admin.entities.appUser.fields.storyMode',
      'colClass' => 'col-xs-4 pad-all',
      'template' => 'WebsiteBundle:Dashboard\\AppUser:storyModeElement.html.twig',
    ]),
    new TextElement([], [
      'value' => $user->getCurrentTier() + 1,
      'label' => 'admin.entities.appUser.fields.currentTier',
      'colClass' => 'col-xs-4 pad-all',
    ]),
    new TextElement([], [
      'value' => $user->getRankScore(),
      'label' => 'admin.entities.appUser.fields.rankScore',
      'colClass' => 'col-xs-4 pad-all',
    ]),
    ]
);
```

#### PanelContainer

Renders the children in a panel.

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard/Containers:panelContainer.html.twig',
      'class' => 'panel'
    ]);
  }
```

Example:
```php
new PanelContainer(
    new AlertElement([], [
      'message' => 'admin.errors.notEnoughDataToRenderGraph'
    ])
);
```

#### RowContainer

Renders the children in a bootstrap row. A row container can be given a `defaultColClass`. Each of its children can be provided with an option `colClass` this is css class that should be applied to the child's column element (defaults to the `defaultColClass` ).

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard/Containers:rowContainer.html.twig',
      'class' => 'row',
      'defaultColClass' => 'col-md-4'
    ]);
  }
```

If the child of a `RowContainer` is given the option `colClass` then the row container will use this as the boostrap column class when rendering the child. If a `defaultColClass` is defined and a child provides a `colClass` than the `colClass` of the child is used when rendering of the element.

Example:
```php
new RowContainer(
  new GraphElement([], [
      'type' => GraphElement::TYPE_RADAR,
      'options' => [
        'scale' => [
          'ticks' => [
            'max' => 100,
          ],
        ],
        'legend' => [
          'display' => false
        ],
        'title' => [
          'display' => true,
          'text' => $graphTitle
        ]
      ],
      'data' => $dataDefinition,
      'colClass' => 'col-md-6'
    ]
  )
);
```

#### TabbedContainer

Renders its children in a tabbed container, each child is given a tabbed view.

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard/Containers:tabbedContainer.html.twig',
      'class' => 'panel'
    ]);
  }
```

Children of a `TabbedContainer` can be given the option `NavName` which is the label that will be used to render the navigation element for that child.

Example:
```php
$panels = new ErrorElement([], [
  'label' => 'something something darkside',
  'navName' => 'A child!'
]);

$element = new TabbedContainer($panels, [
  'headerLabel' => 'admin.entities.appUser.views.stats.miniGameStats',
]);
```

#### TabbedPanelContainer

Render its children in tabbed panel container, each child is given a tabbed view.

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard/Containers:tabbedPanelContainer.html.twig',
      'class' => 'panel',
      'headerLabel' => 'a label'
    ]);
  }
```

Example:
```php
new TabbedPanelContainer($children, ['headerLabel' => 'admin.views.generalStats.top20Rank']);
```

#### TableContainer

Renders its children in a table view

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard/Containers:tableContainer.html.twig',
      'class' => 'table',
      'tableHeading' => [],
      'tableData' => []
    ]);
  }
```

Example:
```php

$data = [
  [
    'myUsername',
    1,
    12312,
    3
  ]
];

$element = new TableContainer([], [
  'tableHeading' => [
    'admin.entities.appUser.fields.username',
    'admin.entities.appUser.fields.rank',
    'admin.entities.appUser.fields.rankScore',
    'admin.entities.appUser.fields.currentTier',

  ],
  'tableData' => $data,
  'navName' => $storyModeName,
  ]
);
```

#### WellContainer

Renders its children in a well.

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'roles' => ['ROLE_ADMIN'],
      'template' => 'BazookasAdminBundle:Dashboard/Containers:wellContainer.html.twig',
      'class' => 'well'
    ]);
  }
```

Example:
```php
$item = new TextElement([], [
  'label' => 'some text'
])

new WellContainer($item);
```

#### FlippablePanelContainer

This container accepts 2 children, a button allows the user to switch between the children. Uses an animation to carry out the transition. This container was initially designed to be used together with the [GraphElement](#graphelement) as the first child and the [StatisticOverviewElement](#statisticoverviewelement) as the second child. Its default icons are therefore a pie chart icon and a table icon. While these are the default states it is perfectly possible to use the container with any of the above elements.

Can be given the options:
```php
  protected function getDefaultOptions() {
    $defaults = parent::getDefaultOptions();
    return array_merge($defaults, [
      'template' => 'BazookasAdminBundle:Dashboard/Containers:flippablePanelContainer.html.twig', //the template to be used by the widget
      'class' => '',  //a class on the parent element of the widget
      'label' => '',  //A panel header label
      'animationClass' => 'flip', // see _animations.scss for available animation styles
      'icon' => 'fa-pie-chart', //The default button icon
      'iconClass' => 'fa',  //the default button's icon class
      'icon2' => 'fa-list', //the icon of the button when its been pushed once by the user
      'iconClass2' => 'fa'  //the second button's icon class
    ]);
  }
```

Example:

```php
   $panel1 = new ErrorElement([], [
      'label' => 'the first panel'
   ]);
   $panel2 = new ErrorElement([], [
      'label' => 'the second panel'
   ]);

   $flipPanel = new FlippablePanelContainer([$panel1, $panel2], [
      'label' => $options['panelLabel']
   ]);
```
