<?php

namespace Bazookas\AdminBundle\Services\Export;

use Doctrine\ORM\EntityManagerInterface;
use Symfony\Component\EventDispatcher\EventDispatcher;
use Symfony\Component\HttpFoundation\StreamedResponse;
use Bazookas\AdminBundle\Event\AdminListEvent;
use Symfony\Component\Translation\TranslatorInterface;

/**
 * @deprecated use BazookasExportBundle instead
 * Class ExportService
 * @package Bazookas\AdminBundle\Services\Export
 */
class ExportService {

  /**
   * @var EntityManagerInterface
   */
  private $entityManager;
  /**
   * @var EventDispatcher
   */
  protected $dispatcher;

  /**
   * @var TranslatorInterface
   */
  protected $translator;

  public function __construct(EntityManagerInterface $entityManager, $eventDispatcher, $translator)
  {
    $this->entityManager = $entityManager;
    $this->dispatcher = $eventDispatcher;
    $this->translator = $translator;
  }

  /**
   * Use function in the following way:
   *
   * $this->exportConfig = [
   *   'entity' => 'CommonBundle:AppUser',
   *   'fields' => [
   *     'facebookId' => 'Some translation',
   *     'profile.firstName' => 'Another translation',
   *     'profile.lastName' => 'Third translation',
   *     'app.id' => false, // don't export the field, just make sure all joins are there
   *   ],
   *   'groupBy' => 'e_options_answers.id',
   *   'fileName' => 'export.csv',
   *   'conditions => [
   *     'e_app.id = :id',
   *     \Doctrine\ORM\Query\Expr\Orx(['e_profile.email IS NOT NULL', 'e_options_answers_user_profile.gender like :gender']),
   *   ],
   *   'parameters => [
   *     'id' => $someId,
   *     'gender' => 'male',
   *   ],
   *  ]);
   * return parent::exportAction($request);
   */

  public function export($exportData, AdminListEvent $event = null) {
    // don't do everything in the streamed response, this way we get a little more error handling
    $selectArray = [];
    $joinArray = [];
    $this->convertFields($exportData['fields'], $joinArray, $selectArray);
    $selectString = implode(', ', $selectArray);

    $qb = $this->entityManager->createQueryBuilder();
    $qb
      ->select($selectString)
      ->from($exportData['entity'], 'e')
    ;
    foreach ($joinArray as $join => $alias) {
      $qb->join($join, $alias);
    }

    // set requirements if necessary
    if (isset($exportData['conditions'])) {
      foreach ($exportData['conditions'] as $condition) {
        $qb->andWhere($condition);
      }
    }
    if (isset($exportData['parameters'])) {
      $qb->setParameters($exportData['parameters']);
    }

    if (isset($event)) {
      $event->queryBuilder = $qb;
      $this->dispatcher->dispatch(AdminListEvent::BEFORE_EXPORT_QUERY, $event);
    }

    if (isset($exportData['groupBy'])) {
      $qb
        ->groupBy($exportData['groupBy'])
      ;
    }

    // get scalar result so we don't need to loop arrays recursively
    $results = $qb->getQuery()->getScalarResult();


    // we've got the results, now it's time to stream the response to the client
    $response = new StreamedResponse();
    $response->setCallback(function() use ($exportData, $results){

      $handle = fopen('php://output', 'w+');
      $headers = array_values($exportData['fields']);

      fputcsv($handle, $headers, ';');

      foreach ($results as $row) {
        $csvData = $this->createCsvDataArray(array_keys($exportData['fields']), $row);
        fputcsv($handle, $csvData, ';');
      }
      fclose($handle);
    });

    $response->setStatusCode(200);
    $response->headers->set('Content-Type', 'text/csv; charset=utf-8');
    $response->headers->set('Content-Disposition', 'attachment; filename="'.$exportData['fileName'].'"');
    return $response;

  }

  private function convertFields(&$original, &$joinArray, &$selectArray) {
    // make sure the root entity is selected
    $tempSelectArray = [
      'e' => ['id']
    ];

    foreach ($original as $key => &$settings) {
      // while we are looping, we can translate the field header
      if (is_string($settings)) {
        $settings = $this->translator->trans($settings, [], 'admin');
      }

      // create an exploded array to create all joins and selects
      $exploded = explode('.', $key);

      // the last element is the actual field
      $selector = array_pop($exploded);

      // we start the selects from `e`, this is the default selected entity
      $joinedEntity = 'e';

      if (count($exploded) > 0) {
        foreach ($exploded as $relation) {
          // create a join and alias
          // by using the join as key we prevent duplicates
          $joinArray[$joinedEntity . '.' . $relation] = $joinedEntity . '_' . $relation;

          // remember the alias for select statements
          $joinedEntity .= '_' . $relation;

          // the following line is to prevent entities from being skipped in the select
          // id is the minimum required field for selecting entities
          // TODO make this work for identifiers not called `id`
          if (empty($tempSelectArray[$joinedEntity]) || !in_array('id', $tempSelectArray[$joinedEntity])) {
            $tempSelectArray[$joinedEntity][] = 'id';
          }
        }
      }

      // when "settings" is false, it means we don't need to select the field
      // just make sure everything is joined
      if ($settings === false) {
        unset($original[$key]);
      }
      else {
        $tempSelectArray[$joinedEntity][] = $selector;
      }
    }

    foreach ($tempSelectArray as $alias => $fields) {
      // make sure there are no duplicates
      $fields = array_unique($fields);

      // make sure the required id is added to the select statement
      // TODO make this work for identifiers not called `id`
      if (array_search('id', $fields) === false) {
        array_unshift($fields, 'id');
      }
      $selectArray[] = 'partial ' . $alias . '.{' . implode(',',$fields) . '}';
    }
  }

  private function createCsvDataArray($keys, $row){
    $csvData = [];
    foreach ($keys as $key){
      // convert the key to the key in the (scalar) query result
      $key = 'e_' . str_replace('.', '_', $key);
      $csvData[] = $row[$key];
    }
    return $csvData;
  }

}
