<?php

namespace Bazookas\AdminBundle\Services\Translation;

use Bazookas\APIFrameworkBundle\Services\Data\Interfaces\DataServiceInterface;
use Symfony\Bundle\FrameworkBundle\Console\Application;
use Symfony\Component\Console\Input\ArrayInput;
use Symfony\Component\Console\Output\BufferedOutput;
use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\HttpKernel\KernelInterface;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;
use Symfony\Component\Translation\Translator;
use Symfony\Component\Translation\TranslatorInterface;
use Symfony\Component\Yaml\Yaml;

class TranslationDataService implements DataServiceInterface
{

  /**
   * @var Translator
   */
  private $translator;

  /**
   * @var string
   */
  private $kernelRootDir;

  /**
   * @var string
   */
  private $kernelCacheDir;

  /**
   * @var KernelInterface
   */
  private $kernel;

  /**
   * TranslationDataService constructor.
   * @param TranslatorInterface $translator
   * @param $kernelRootDir
   * @param $kernelCacheDir
   * @param KernelInterface $kernel
   */
  public function __construct(TranslatorInterface $translator, $kernelRootDir, $kernelCacheDir, KernelInterface $kernel) {
    $this->translator = $translator;
    $this->kernelRootDir = $kernelRootDir;
    $this->kernelCacheDir = $kernelCacheDir;
    $this->kernel = $kernel;
  }

  /**
   * @param $data
   * @param array $translationDomains
   */
  public function updateTranslation($data, array $translationDomains)
  {
    list($domain, $key) = explode(':', $data['id']);

    $allowed = false;
    foreach ($translationDomains as $translationDomain) {
      if ($translationDomain == $domain) {
        $allowed = true;
      }
    }

    if (!$allowed) {
      throw new AccessDeniedException();
    }

    $locale = $data['field'];

    $translationCatalogue = $this->translator->getCatalogue($locale)->all($domain);

    //If the value hasn't changed then don't bother doing anything
    if (!isset($translationCatalogue[$key]) || $data['value'] !== $translationCatalogue[$key]) {
      $translationCatalogue[$key] = $data['value'];

      $translationCatalogue = $this->expandToAssocArray($translationCatalogue);

      list($directory, $fileName) = $this->getFilePath($domain, $locale);
      $this->writeToFile($directory, $fileName, $translationCatalogue);
    }
  }

  /**
   * Expand an input array with dot seperated keys into a multi-level associative array. For example ["an.example" => "data]
   * becomes [ "an" => [ "example" => "data" ] ]
   * @param array $data
   * @return array
   */
  private function expandToAssocArray(array $data) {
    $expandedData = [];
    foreach($data as $itemKey => $item) {
      $keys = explode('.', $itemKey);
      $arr = &$expandedData;

      for ($i =0; $i < count($keys); $i++) {
        if (!isset($arr[$keys[$i]])) {
          if ($i + 1 == count($keys)) {
            $arr[$keys[$i]] = $item;
          } else {
            $arr[$keys[$i]] = [];
          }
        }

        $arr = &$arr[$keys[$i]];
      }
    }

    return $expandedData;
  }

  /**
   * @param $directory
   * @param $fileName
   * @param array $data
   */
  private function writeToFile($directory, $fileName, array $data) {
    $yaml = Yaml::dump($data, 8);

    $fs = new Filesystem();
    if (!$fs->exists($directory)) {
      $fs->mkdir($directory, 0777);
    }

    $fileExists = $fs->exists($directory .'/'.$fileName);
    file_put_contents($directory .'/'.$fileName, $yaml);

    $this->clearCache($fileExists);
  }

  /**
   * @param $domain
   * @param $locale
   * @return array[$directory, $fileName]
   */
  private function getFilePath($domain, $locale) {
    return [$this->kernelRootDir . '/Resources/translations', $domain . '.' .$locale . '.yml'];
  }

  /**
   * @param bool $fileExists did the file exist previously
   */
  private function clearCache($fileExists) {
    $cacheDir = dirname($this->kernelCacheDir);

    foreach (['prod', 'dev', 'test'] as $env) {
      if (!empty($cacheDir)) {
        $translationDirectories = glob("$cacheDir/$env/translations/*", GLOB_ERR);
        if (is_array($translationDirectories)) {
          array_map('unlink', $translationDirectories);
        }
      }
    }

    if (!$fileExists) {
      $this->executeClearCacheCommand();
    }
  }

  private function executeClearCacheCommand() {
    $kernel = $this->kernel;
    $application = new Application($kernel);
    $application->setAutoExit(false);

    $input = new ArrayInput(array(
      'command' => 'cache:clear',
      '--env' => $kernel->getEnvironment(),
      '--no-warmup' => true
    ));

    $output = new BufferedOutput();
    $application->run($input, $output);
  }
}
