<?php

namespace Bazookas\AdminBundle\Twig\Base;

use Bazookas\AdminBundle\AdminElements\Containers\MenuContainer;
use Bazookas\AdminBundle\AdminElements\Elements\Actions\MenuActions\MenuActionElement;
use Bazookas\AdminBundle\AdminElements\Elements\Actions\MenuActions\MenuHeaderElement;
use Bazookas\AdminBundle\AdminElements\Interfaces\AdminElementInterface;
use Bazookas\CommonBundle\Security\Roles;
use Symfony\Component\HttpFoundation\Request;

class BaseAdminMenuExtension extends \Twig_Extension
{

  /**
   * @var AdminElementInterface[]
   */
  protected $elements = [];

  /**
   * @var string[]
   */
  protected $bundles;

  /**
   * @deprecated
   * @var array
   */
  protected $v1MenuConfig;

  public function __construct($bundles, $v1MenuConfig) {
    $this->bundles = $bundles;
    $this->v1MenuConfig = $v1MenuConfig;
  }

  /**
   * Get Twig functions defined in this extension.
   *
   * @return array
   */
  public function getFunctions()
  {
    return array(
      new \Twig_SimpleFunction('buildAdminMenu', array($this, 'buildAdminMenu'), [
        'needs_environment' => true
      ]),
    );
  }

  /**
   * @param \Twig_Environment $env
   * @return AdminElementInterface[]
   */
  public function buildAdminMenu(\Twig_Environment $env) {
    $app = $env->getGlobals()['app'];

    $request = $app->getRequest();
    $this->createMenu($request);

    return $this->elements;
  }

  /**
   * @param Request $request
   */
  protected function createMenu(Request $request) {
    $this->addMenuHeader();
    $this->addAdminBundleMenuItems();
    $this->addMediaBundleMenuItems();
    $this->addCronBundleMenuItems();
  }

  protected function addMenuHeader() {
    $this->elements[] = new MenuHeaderElement([], [
      'label' => 'admin.menu.admin.header'
    ]);
  }

  /**
   * Create and add the admin bundle menu items
   */
  protected function addAdminBundleMenuItems() {
    //Add the user management menu items
    $this->elements[] = new MenuContainer([
      new MenuActionElement([], [
        'label' => 'admin.entities.cMSUser.namePlural',
        'route' => 'admin_cms_user',
        'iconClass' => 'ti ti-user',
        'roles' => [Roles::ROLE_ADD_USER],
      ]),
      new MenuActionElement([], [
        'label' => 'admin.entities.cMSUserGroup.namePlural',
        'route' => 'admin_cms_user_group',
        'iconClass' => 'ti-id-badge',
        'roles' => [Roles::ROLE_ADD_GROUP],
      ])
    ],
    [
      'label' => 'admin.menu.users',
      'roles' => [Roles::ROLE_ADD_USER, Roles::ROLE_ADD_GROUP],
      'iconClass' => 'ti ti-user'
    ]);

    //Add the copy module menu items
    $this->elements[] = new MenuActionElement([], [
      'label' => 'admin.menu.copy.label',
      'route' => 'admin_copy',
      'iconClass' => 'fa fa-language fa-3x',
      'roles' => [ Roles::ROLE_COPY_MANAGER ],
    ]);
  }

  protected function addMediaBundleMenuItems() {
    if (isset($this->bundles['BazookasMediaBundle'])) {
      $this->elements[] = new MenuContainer([
        new MenuActionElement([], [
          'label' => 'media.entities.image.namePlural',
          'route' => 'bazookas_media_image_admin',
          'icon' => 'ti-image',
          'roles' => [Roles::ROLE_MEDIA_ADMIN],
        ]),
        new MenuActionElement([], [
          'label' => 'media.entities.video.namePlural',
          'route' => 'bazookas_media_video_admin',
          'iconClass' => 'ti-video-clapper',
          'roles' => [Roles::ROLE_MEDIA_ADMIN],
        ]),
        new MenuActionElement([], [
          'label' => 'media.entities.videoProvider.namePlural',
          'route' => 'bazookas_media_video_provider_admin',
          'icon' => 'ti-settings',
          'roles' => [Roles::ROLE_MEDIA_ADMIN],
        ]),
        new MenuActionElement([], [
          'label' => 'media.entities.mediaFolder.namePlural',
          'route' => 'bazookas_media_media_folder_admin',
          'iconClass' => 'ti-folder',
          'roles' => [Roles::ROLE_MEDIA_ADMIN],
        ])
      ],
      [
        'label' => 'media.menu.media',
        'roles' => [Roles::ROLE_MEDIA_ADMIN],
        'icon' => 'ti-image'
      ]);
    }
  }

  protected function addCronBundleMenuItems() {
    if (isset($this->bundles['BazookasCronBundle'])) {
      $this->elements[] = new MenuActionElement([], [
        'label' => 'admin.entities.cron.namePlural',
        'route' => 'bazookas_cron_cron_admin',
        'iconClass' => 'ti-list',
        'roles' => [Roles::ROLE_CRON_ADMIN],
      ]);
    }
  }

  /**
   * @deprecated to be removed in v2
   * This function adds backwards compatibility between the v1 and v2 menu structures.
   */
  protected function addV1MenuItems() {
    foreach($this->v1MenuConfig['main'] as $key => $config) {
      if ($config['type'] == 'submenu') {
        $element = new MenuContainer();

        foreach($config['children'] as $childConfig) {
          $childElement = new MenuActionElement();

          $childElement->setOptions([
            'label' => $childConfig['label'],
            'route' => $childConfig['route'],
            'iconClass' => $childConfig['icon'],
            'roles' => $childConfig['roles']
          ]);

          $element->addChild($childElement);
        }
      } else {
        $element = new MenuActionElement();
      }

      $options = [
        'label' => $config['label'],
        'iconClass' => $config['icon'],
        'roles' => $config['roles']
      ];

      if ($element instanceof MenuActionElement) {
        $options['route'] = $config['route'];
      }

      $element->setOptions($options);

      $this->addV1Element($element);
    }
  }

  /**
   * @deprecated to be removed in v2
   * @param AdminElementInterface $v1Element
   */
  private function addV1Element(AdminElementInterface $v1Element) {
    $addElement = true;
    foreach($this->elements as $element) {
      if ($v1Element instanceof MenuContainer) {
        if ($element instanceof MenuContainer && $v1Element->getLabel() == $element->getLabel()) {
          $addElement = false;
          break;
        }
      } else if ($element instanceof MenuActionElement && $element->getRoute() == $v1Element->getRoute()) {
        $addElement = false;
        break;
      }
    }

    if ($addElement) {
      $this->elements[] = $v1Element;
    }

  }
}
