<?php

namespace WebsiteBundle\Controller\Admin;

use Bazookas\AdminBundle\AdminElements\Elements\Actions\PageActions\ExportPageActionElement;
use Bazookas\AdminBundle\Controller\Base\BaseAdminListController;
use Bazookas\AdminBundle\PageBuilder\Base\BaseFormPageBuilder;
use Bazookas\AdminBundle\PageBuilder\BulkEditPageBuilder;
use Bazookas\AdminBundle\PageBuilder\Interfaces\ListPageBuilderInterface;
use Bazookas\AdminBundle\PageBuilder\ListPageBuilder;
use Bazookas\ExportBundle\Services\Interfaces\ExportServiceInterface;
use Symfony\Component\HttpFoundation\Request;
use WebsiteBundle\Entity\Experiment;
use WebsiteBundle\Entity\Participant;
use WebsiteBundle\Form\Admin\ParticipantAdminType;
use WebsiteBundle\Form\Admin\ParticipantBulkEditAdminType;
use WebsiteBundle\PageBuilder\ParticipantViewPageBuilder;
use WebsiteBundle\Security\Roles;

class ParticipantsAdminController extends BaseAdminListController
{
  public function __construct() {
    parent::__construct();

    $this->builders[BaseAdminListController::ACTION_ADD] = '@WebsiteBundle.PageBuilder.ParticipantFormPageBuilder';
    $this->builders[BaseAdminListController::ACTION_EDIT] = '@WebsiteBundle.PageBuilder.ParticipantFormPageBuilder';
    $this->builders[BaseAdminListController::ACTION_VIEW] = ParticipantViewPageBuilder::class;
  }

  protected function getEntityClass()
  {
    return Participant::class;
  }

  protected function modifyListBuilder(Request $request, ListPageBuilderInterface $builder)
  {
    /**
     * @var ListPageBuilder $builder
     */
    $builder = parent::modifyListBuilder($request, $builder);

    //Add fields
    $builder
      ->addField('id')
      ->addBooleanField('active')
      ->addField('experiment', [
        'template' => 'WebsiteBundle:List\Fields:experimentField.html.twig'
      ])
      ->addField('sessions', [
        'label' => 'admin.entities.testSession.fields.number',
        'template' => 'WebsiteBundle:List\Fields:sessionNumber.html.twig'
      ])
      ->addField('condition', [
        'template' => 'WebsiteBundle:List\Fields:condition.html.twig'
      ])
      ->addField('name', [
        'roles' => ['ROLE_ADMIN']
      ])
      ->addField('surname', [
        'roles' => ['ROLE_ADMIN']
      ])
      ->addField('email', [
        'roles' => ['ROLE_ADMIN']
      ])
      ->addField('sessions', [
        'label' => 'admin.entities.participant.fields.scheduled',
        'template' => 'WebsiteBundle:List\Fields:sessionScheduled.html.twig'
      ])
      ->addField('sessions', [
        'label' => 'admin.entities.testSession.fields.completed',
        'template' => 'WebsiteBundle:List\Fields:sessionCompleted.html.twig'
      ])
    ;

    //Add head researcher filters
    if ($this->isGranted('ROLE_ADMIN_RESEARCHER')) {
      $builder
        ->addTextFilterField('name')
        ->addTextFilterField('surname')
        ->addTextFilterField('email')
      ;
    }

    //add general filters
    $builder
      ->addBooleanFilterField('active')
      ->addTextFilterField('experiment.name', [
        'label' => 'admin.entities.participant.fields.experiment'
      ])
      ->addTextFilterField('sessions.number', [
        'label' => 'admin.entities.testSession.fields.number',
      ])

      ->setMaintainFilterState(true)
    ;

    //Add the export action
    if ($this->hasAccess(self::ACTION_EXPORT)) {
      $builder->prependAction(new ExportPageActionElement([], [
        'route' => $request->get('_route'),
        'action' => self::ACTION_EXPORT,
      ]));
    }

    //Pass the user to the list query builder
    $builder->setListQueryParams([
      'isAdminResearcher' => $this->isGranted(Roles::ROLE_ADMIN_RESEARCHER),
      'user' => $this->getUser()
    ]);

    return $builder;
  }

  /**
   * @param $action
   * @return null
   */
  protected function getFormClass($action) {
    switch($action) {
      case self::ACTION_ADD:
      case self::ACTION_EDIT:
        return ParticipantAdminType::class;
      case self::ACTION_BULK_EDIT:
        return ParticipantBulkEditAdminType::class;
      default:
        return null;
    }
  }

  /**
   * @param string $action
   * @return bool
   */
  protected function hasAccess($action) {
    switch($action) {
      case self::ACTION_LIST:
      case self::ACTION_ADD:
      case self::ACTION_VIEW:
        return true;
      case self::ACTION_BULK_EDIT:
      case self::ACTION_EDIT:
      case self::ACTION_REMOVE:
      case self::ACTION_EXPORT:
        return $this->isGranted(Roles::ROLE_ADMIN_RESEARCHER);
      default:
        return false;
    }
  }

  /**
   * @param Request $request
   * @param BaseFormPageBuilder $builder
   * @return BaseFormPageBuilder
   */
  protected function modifyAddBuilder(Request $request, BaseFormPageBuilder $builder)
  {
    $formOptions = $this->getFormOptions();
    $builder->setFormOptions($formOptions);
    return parent::modifyAddBuilder($request, $builder);
  }

  /**
   * @param Request $request
   * @param BaseFormPageBuilder $builder
   * @return BaseFormPageBuilder
   */
  protected function modifyEditBuilder(Request $request, BaseFormPageBuilder $builder)
  {
    $formOptions = $this->getFormOptions();
    $builder->setFormOptions($formOptions);
    return parent::modifyEditBuilder($request, $builder);
  }

  /**
   * @param Request $request
   * @param BulkEditPageBuilder $builder
   * @return BulkEditPageBuilder
   */
  protected function modifyBulkEditBuilder(Request $request, BulkEditPageBuilder $builder)
  {
    $formOptions = $this->getFormOptions();
    $builder->setFormOptions($formOptions);
    return parent::modifyBulkEditBuilder($request, $builder);
  }

  /**
   * @return array
   */
  private function getFormOptions() {
    $experimentRepo = $this->get('doctrine.orm.entity_manager')->getRepository(Experiment::class);
    $experiments = $experimentRepo->getAllExperimentsForUser($this->isGranted(Roles::ROLE_ADMIN_RESEARCHER), $this->getUser());

    return [
      'experiments' => $experiments
    ];
  }

  /**
   * @param Request $request
   * @param $action
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   */
  protected function viewAction(Request $request, $action) {
    $testSessionController = new TestSessionAdminController();
    $testSessionController->setContainer($this->container);

    $participantId = $request->get('_id');
    $request->query->add(['participantId' => $participantId]);

    return $testSessionController->indexAction($request, self::ACTION_LIST);
  }

  /**
   * @param Request $request
   * @param $action
   * @return \Symfony\Component\HttpFoundation\Response
   * @throws \Exception
   */
  protected function exportAction(Request $request, $action) {
    $config = [
      'sheets' => [
        [
          'entity' => Participant::class,
          'fields' => [
            'id' => 'admin.entities.participant.fields.id',
            'name'  => 'admin.entities.participant.fields.name',
            'surname'  => 'admin.entities.participant.fields.surname',
            'email'  => 'admin.entities.participant.fields.email',
            'agreesToTermsAndConditions'  => 'admin.entities.participant.fields.agreesToTermsAndConditions',
            'experiment.name'  => 'admin.entities.participant.fields.experiment',
            'active'  => 'admin.entities.participant.fields.active',
            'condition.testGroup' => 'admin.entities.participant.fields.testGroup',
            'condition.gamificationLevel' => 'admin.entities.participant.fields.gamificationLevel',
          ]
        ]
      ],
      'fileName' => 'participants-' . date('Y-m-d')
    ];

    $exportService = $this->get('bazookas.export.service.ExportService');

    if (!$exportService instanceof ExportServiceInterface) {
      throw new \Exception('Invalid export service, must implement ' . ExportServiceInterface::class);
    }

    return $exportService->export($request, $config);
  }
}
