<?php

namespace WebsiteBundle\Entity;

use Bazookas\CommonBundle\Entity\Base\BaseCMSUser;
use Bazookas\CommonBundle\Entity\Base\BaseEntity;
use Doctrine\Common\Collections\ArrayCollection;
use Doctrine\ORM\Mapping as ORM;
use Symfony\Component\Security\Core\Authorization\AuthorizationChecker;
use Symfony\Component\Validator\Constraints as Assert;
use Symfony\Component\Validator\Context\ExecutionContextInterface;
use WebsiteBundle\Security\Roles;

/**
 * @ORM\Entity(repositoryClass="WebsiteBundle\Repository\ExperimentRepository")
 * @ORM\HasLifecycleCallbacks
 */
class Experiment extends BaseEntity {

  //Constants
  const SESSION_LOCATION_LAB = 'lab';
  const SESSION_LOCATION_HOME = 'home';

  /**
   * @ORM\Column(type="string")
   * @var string
   * @Assert\NotBlank
   */
  private $name;

  /**
   * @ORM\ManyToMany(targetEntity="Researcher", inversedBy="experiments")
   * @var ArrayCollection
   */
  private $researchers;

  /**
   * @ORM\OneToMany(targetEntity="Participant", mappedBy="experiment")
   * @var ArrayCollection
   */
  private $participants;

  /**
   * @ORM\OneToMany(targetEntity="PsychEducationPanel", mappedBy="experiment", cascade="all", orphanRemoval=true)
   * @Assert\Valid
   */
  private $psychEducationPanels;

  /**
   * @ORM\Column(type="integer")
   * @var integer
   * @Assert\NotBlank
   * @Assert\GreaterThan(2)
   */
  private $totalSessions;

  /**
   * @ORM\Column(type="integer")
   * @var integer (in seconds)
   * @Assert\NotBlank
   */
  private $sessionDuration;

  /**
   * @ORM\Column(type="integer")
   * @var integer (in minutes)
   * @Assert\NotBlank
   */
  private $minTimeBetweenSession;

  /**
   * @ORM\Column(type="integer")
   * @var integer (in minutes)
   * @Assert\NotBlank
   */
  private $maxTimeBetweenSession;

  /**
   * @ORM\Column(type="string")
   * @var string
   * @Assert\Choice(choices={"lab", "home"})
   */
  private $firstSessionLocation = self::SESSION_LOCATION_LAB;

  /**
   * @ORM\Column(type="string")
   * @var string
   * @Assert\Choice(choices={"lab", "home"})
   */
  private $lastSessionLocation = self::SESSION_LOCATION_LAB;

  /**
   * @ORM\Column(type="integer")
   * @var integer (in days)
   * @Assert\NotBlank
   */
  private $experimentDuration;

  /**
   * @ORM\Column(type="integer")
   * @var integer (in minutes)
   * @Assert\NotBlank
   */
  private $notificationTimeBeforeSession = 60;

  /**
   * @ORM\Column(type="boolean")
   * @var bool
   */
  private $notifyBeforeSession = true;

  /**
   * @ORM\Column(type="integer")
   * @var integer (in minutes)
   * @Assert\NotBlank
   */
  private $notificationTimeAfterSession = 60;

  /**
   * @ORM\Column(type="boolean")
   * @var bool
   */
  private $notifyAfterSession = true;

  /**
   * @ORM\OneToMany(targetEntity="MoodQuestion", mappedBy="experiment", cascade={"all"})
   */
  private $moodQuestions;

  /**
   * @ORM\OneToMany(targetEntity="TestSession", mappedBy="experiment")
   * @var TestSession[]
   */
  private $sessions;

  /**
   * @ORM\Column(type="integer")
   * @var int
   */
  private $participantCount = 0;

  /**
   * @ORM\Column(type="integer")
   * @var int
   */
  private $activeParticipantCount = 0;

  /**
   * The number of completed sessions
   * @ORM\Column(type="integer")
   * @var integer)
   * @Assert\NotBlank
   */
  private $sessionCount = 0;

  /**
   * @ORM\ManyToMany(targetEntity="WebsiteBundle\Entity\Quote")
   * @var Quote[]
   */
  private $availableQuotes;

  /**
   * @ORM\OneToMany(targetEntity="ExperimentCondition", mappedBy="experiment", cascade={"all"})
   * @var ExperimentCondition[]
   * @Assert\Count(min=1, max=4, minMessage="admin.entities.experiment.minExperimentConditions", maxMessage="admin.entities.experiment.maxExperimentConditions")
   */
  private $conditions;

  public function __construct() {
    parent::__construct();
    $this->participants = new ArrayCollection();
    $this->researchers = new ArrayCollection();
    $this->psychEducationPanels = new ArrayCollection();
    $this->moodQuestions = new ArrayCollection();
    $this->conditions = new ArrayCollection();
  }

  /**
   * @return string|null
   */
  public function getName()
  {
    return $this->name;
  }

  /**
   * @param string $name
   * @return $this
   */
  public function setName($name)
  {
    $this->name = $name;

    return $this;
  }

  /**
   * @return ArrayCollection
   */
  public function getResearchers()
  {
    return $this->researchers;
  }

  /**
   * @param Researcher $researchers
   * @return $this
   */
  public function setResearchers(Researcher $researchers)
  {
    $this->researchers = $researchers;

    return $this;
  }

  /**
   * @param Researcher $researcher
   * @return $this
   */
  public function addResearcher(Researcher $researcher) {
    $this->researchers->add($researcher);
    return $this;
  }

  /**
   * @param Researcher $researcher
   * @return $this
   */
  public function removeResearcher(Researcher $researcher) {
    $this->researchers->removeElement($researcher);
    return $this;
  }

  /**
   * @return ArrayCollection
   */
  public function getParticipants()
  {
    return $this->participants;
  }

  /**
   * @param ArrayCollection|Participant[] $participants
   * @return $this
   */
  public function setParticipants($participants)
  {
    if (is_array($participants)) {
      $this->participants = new ArrayCollection($participants);
    } else {
      $this->participants = $participants;
    }

    return $this;
  }

  /**
   * @return ArrayCollection
   */
  public function getPsychEducationPanels()
  {
    return $this->psychEducationPanels;
  }

  /**
   * @param ArrayCollection|PsychEducationPanel[] $psychEducationPanels
   * @return $this
   */
  public function setPsychEducationPanels($psychEducationPanels)
  {
    if (is_array($psychEducationPanels) && !empty($psychEducationPanels)) {
      $psychEducationPanels = new ArrayCollection($psychEducationPanels);
    }

    $this->psychEducationPanels = $psychEducationPanels;

    return $this;
  }

  /**
   * @param PsychEducationPanel $psychEducationPanel
   * @return $this
   */
  public function addPsychEducationPanel(PsychEducationPanel $psychEducationPanel) {
    $psychEducationPanel->setExperiment($this);
    $this->psychEducationPanels->add($psychEducationPanel);
    return $this;
  }

  /**
   * @param PsychEducationPanel $psychEducationPanel
   * @return $this
   */
  public function removePsychEducationPanel(PsychEducationPanel $psychEducationPanel) {
    $this->psychEducationPanels->removeElement($psychEducationPanel);
    return $this;
  }

  /**
   * @return int
   */
  public function getSessionCount()
  {
    return $this->sessionCount;
  }

  /**
   * @param int $sessionCount
   * @return $this
   */
  public function setSessionCount($sessionCount)
  {
    $this->sessionCount = $sessionCount;

    return $this;
  }

  /**
   * @return int
   */
  public function getSessionDuration()
  {
    return $this->sessionDuration;
  }

  /**
   * @param int $sessionDuration
   * @return $this
   */
  public function setSessionDuration($sessionDuration)
  {
    $this->sessionDuration = $sessionDuration;

    return $this;
  }

  /**
   * @return int
   */
  public function getMinTimeBetweenSession()
  {
    return $this->minTimeBetweenSession;
  }

  /**
   * @param int $minTimeBetweenSession
   * @return $this
   */
  public function setMinTimeBetweenSession($minTimeBetweenSession)
  {
    $this->minTimeBetweenSession = $minTimeBetweenSession;

    return $this;
  }

  /**
   * @return int
   */
  public function getMaxTimeBetweenSession()
  {
    return $this->maxTimeBetweenSession;
  }

  /**
   * @param int $maxTimeBetweenSession
   * @return $this
   */
  public function setMaxTimeBetweenSession($maxTimeBetweenSession)
  {
    $this->maxTimeBetweenSession = $maxTimeBetweenSession;

    return $this;
  }

  /**
   * @return string
   */
  public function getFirstSessionLocation()
  {
    return $this->firstSessionLocation;
  }

  /**
   * @param string $firstSessionLocation
   * @return Experiment
   */
  public function setFirstSessionLocation($firstSessionLocation)
  {
    $this->firstSessionLocation = $firstSessionLocation;

    return $this;
  }

  /**
   * @return string
   */
  public function getLastSessionLocation()
  {
    return $this->lastSessionLocation;
  }

  /**
   * @param string $lastSessionLocation
   * @return $this
   */
  public function setLastSessionLocation($lastSessionLocation)
  {
    $this->lastSessionLocation = $lastSessionLocation;

    return $this;
  }

  /**
   * @return int
   */
  public function getExperimentDuration()
  {
    return $this->experimentDuration;
  }

  /**
   * @param int $experimentDuration
   * @return $this
   */
  public function setExperimentDuration($experimentDuration)
  {
    $this->experimentDuration = $experimentDuration;

    return $this;
  }

  /**
   * @return int
   */
  public function getNotificationTimeBeforeSession()
  {
    return $this->notificationTimeBeforeSession;
  }

  /**
   * @param int $notificationTimeBeforeSession
   * @return $this
   */
  public function setNotificationTimeBeforeSession($notificationTimeBeforeSession)
  {
    $this->notificationTimeBeforeSession = $notificationTimeBeforeSession;

    return $this;
  }

  /**
   * @return bool
   */
  public function isNotifyBeforeSession()
  {
    return $this->notifyBeforeSession;
  }

  /**
   * @param bool $notifyBeforeSession
   * @return Experiment
   */
  public function setNotifyBeforeSession($notifyBeforeSession)
  {
    $this->notifyBeforeSession = $notifyBeforeSession;

    return $this;
  }

  /**
   * @return int
   */
  public function getNotificationTimeAfterSession()
  {
    return $this->notificationTimeAfterSession;
  }

  /**
   * @param int $notificationTimeAfterSession
   * @return $this
   */
  public function setNotificationTimeAfterSession($notificationTimeAfterSession)
  {
    $this->notificationTimeAfterSession = $notificationTimeAfterSession;

    return $this;
  }

  /**
   * @return bool
   */
  public function isNotifyAfterSession()
  {
    return $this->notifyAfterSession;
  }

  /**
   * @param bool $notifyAfterSession
   * @return Experiment
   */
  public function setNotifyAfterSession($notifyAfterSession)
  {
    $this->notifyAfterSession = $notifyAfterSession;

    return $this;
  }

  /**
   * @return ArrayCollection
   */
  public function getMoodQuestions()
  {
    return $this->moodQuestions;
  }

  /**
   * @param $moodQuestions
   * @return $this
   */
  public function setMoodQuestions($moodQuestions)
  {
    $this->moodQuestions = $moodQuestions;

    return $this;
  }

  /**
   * @param MoodQuestion $question
   * @return $this
   */
  public function addMoodQuestion(MoodQuestion $question) {
    $question->setExperiment($this);
    $this->moodQuestions->add($question);
    return $this;
  }

  /**
   * @param MoodQuestion $question
   * @return $this
   */
  public function removeMoodQuestion(MoodQuestion $question) {
    $this->moodQuestions->removeElement($question);
    return $this;
  }

  /**
   * @return TestSession[]
   */
  public function getSessions()
  {
    return $this->sessions;
  }

  /**
   * @param TestSession[] $sessions
   * @return $this
   */
  public function setSessions($sessions)
  {
    $this->sessions = $sessions;

    return $this;
  }

  /**
   * @return int
   */
  public function getParticipantCount()
  {
    return $this->participantCount;
  }

  /**
   * @param int $participantCount
   * @return $this
   */
  public function setParticipantCount($participantCount)
  {
    $this->participantCount = $participantCount;

    return $this;
  }

  /**
   * @return int
   */
  public function getActiveParticipantCount()
  {
    return $this->activeParticipantCount;
  }

  /**
   * @param int $activeParticipantCount
   * @return $this
   */
  public function setActiveParticipantCount($activeParticipantCount)
  {
    $this->activeParticipantCount = $activeParticipantCount;

    return $this;
  }

  /**
   * @return int
   */
  public function getTotalSessions()
  {
    return $this->totalSessions;
  }

  /**
   * @param int $totalSessions
   * @return $this
   */
  public function setTotalSessions($totalSessions)
  {
    $this->totalSessions = $totalSessions;

    return $this;
  }

  /**
   * @return ExperimentCondition[]
   */
  public function getConditions()
  {
    return $this->conditions;
  }

  /**
   * @param ExperimentCondition[] $conditions
   * @return Experiment
   */
  public function setConditions($conditions)
  {
    $this->conditions = $conditions;

    return $this;
  }

  /**
   * @param ExperimentCondition $condition
   * @return $this
   */
  public function addCondition(ExperimentCondition $condition) {
    $condition->setExperiment($this);
    $this->conditions->add($condition);
    return $this;
  }

  /**
   * @param ExperimentCondition $condition
   * @return $this
   */
  public function removeCondition(ExperimentCondition $condition) {
    $this->conditions->removeElement($condition);
    return $this;
  }

  /**
   * @return Quote[]
   */
  public function getAvailableQuotes()
  {
    return $this->availableQuotes;
  }

  /**
   * @param Quote[] $availableQuotes
   * @return Experiment
   */
  public function setAvailableQuotes($availableQuotes)
  {
    $this->availableQuotes = $availableQuotes;

    return $this;
  }

  /**
   * @param AuthorizationChecker $auth
   * @param BaseCMSUser $user
   * @return bool
   */
  public function canDelete(AuthorizationChecker $auth, BaseCMSUser $user) {
    if (count($this->sessions) > 0) {
      return false;
    } else {
      return $auth->isGranted($this->getRoleRequiredForDelete());
    }
  }

  /**
   * @return string
   */
  public function getRoleRequiredForView() {
    return Roles::ROLE_CMS_USER;
  }

  /**
   * @Assert\Callback
   */
  public function validate(ExecutionContextInterface $context, $payload)
  {
    //Check for duplicate conditions
    $conditions = [];
    foreach($this->conditions as $key => $condition) {
      $hash = md5($condition->getTestGroup().$condition->getGamificationLevel());
      $path = 'form.conditions['.$key.'].condition';

      if (isset($conditions[$hash])) {
        $context->buildViolation('admin.entities.experiment.duplicateCondition')
          ->atPath($path)
          ->setParameters([
            '{{testGroup}}' => $condition->getTestGroup(),
            '{{gamificationLevel}}' => $condition->getGamificationLevel()
          ])
          ->addViolation()
        ;
      } else {
        $conditions[$hash] = $path;
      }
    }
  }
}
