<?php

namespace WebsiteBundle\Repository;

use Bazookas\AdminBundle\Repository\Base\BaseRepository;
use WebsiteBundle\Entity\Experiment;
use WebsiteBundle\Entity\Participant;
use WebsiteBundle\Entity\Researcher;
use WebsiteBundle\Entity\TestSession;
use WebsiteBundle\Entity\UnlockedQuote;
use WebsiteBundle\Security\Roles;

class ParticipantRepository extends BaseRepository
{

  public function getEntityClass() {
    return Participant::class;
  }

  public function createListQueryBuilder(array $fields, array $params) {
    $qb = $this->createQueryBuilder('e');
    $qb
      ->addSelect('sessions')
      ->leftJoin('e.sessions', 'sessions')
      ->addSelect('experiment')
      ->leftJoin('e.experiment', 'experiment')
      ->addSelect('condition')
      ->leftJoin('e.condition', 'condition')
    ;

    //Non admin researcher may only see the experiments they are a part of
    if (!$params['isAdminResearcher']) {
      $user = $params['user'];

      $qb
        ->join('experiment.researchers', 'researchers', 'WITH', $qb->expr()->eq('researchers.id', ':user'))
        ->setParameter('user', $user)
      ;
    }

    $qb
      ->addOrderBy('e.id', 'ASC')
      ->addOrderBy('sessions.number', 'DESC')
    ;

    return $qb;
  }

  public function setParticipantsToInactive(Experiment $experiment) {
    $sessions = $this->getParticipantsToInactive($experiment);

    if (!empty($sessions)) {
      $participantIds = [];
      foreach($sessions as $session) {
        $participantIds[] = $session['participant']['id'];
      }

      $qb = $this->_em->createQueryBuilder();
      $qb
        ->update(Participant::class, 'participant')
        ->set('participant.active', ':false')
        ->andWhere($qb->expr()->in('participant.id', ':ids'))
        ->setParameters([
          'false' => false,
          'ids' => $participantIds
        ])
      ;

      $qb->getQuery()->execute();
    }
  }

  private function getParticipantsToInactive(Experiment $experiment) {
    $qb = $this->createGreatestRowPerGroupQueryBuilder(TestSession::class, 'participant', 'availableFrom');

    //maxTimeBetween = max time + 48 hours
    $maxTimeBetween = $experiment->getMaxTimeBetweenSession() + 2880 - $experiment->getMinTimeBetweenSession();

    $maxTime = new \DateTime('now', new \DateTimeZone('UTC'));
    $maxTime->modify('- '.$maxTimeBetween.' minutes');

    $rootAlias = $qb->getRootAliases()[0];

    $qb
      ->addSelect('participant')
      ->join($rootAlias .'.participant', 'participant', 'WITH', $qb->expr()->eq('participant.active', ':true'))

      ->andWhere($qb->expr()->eq($rootAlias .'.experiment', ':experiment'))
      ->andWhere($qb->expr()->lt($rootAlias .'.availableFrom', ':maxTime'))

      ->andWhere(
        $qb->expr()->orX(
          $qb->expr()->lt('participant.activeOverrideExpirationDate', ':now'),
          $qb->expr()->isNull('participant.activeOverrideExpirationDate')
        ))
      ->setParameters([
        'experiment' => $experiment,
        'maxTime' => $maxTime,
        'true' => true,
        'now' => new \DateTime('now', new \DateTimeZone('UTC'))
      ])
    ;

    return $qb->getQuery()->getArrayResult();
  }

  public function getUnlockedQuotes(Participant $participant) {
    $qb = $this->_em->createQueryBuilder();
    $qb
      ->select('unlockedQuote')
      ->from(UnlockedQuote::class, 'unlockedQuote')

      ->addSelect('quote')
      ->leftJoin('unlockedQuote.quote', 'quote')

      ->addSelect('session')
      ->leftJoin('unlockedQuote.session', 'session')

      ->andWhere($qb->expr()->eq('unlockedQuote.participant', ':participant'))
      ->setParameter('participant', $participant)
    ;

    return $qb->getQuery()->getResult();
  }

}
