<?php

namespace WebsiteBundle\Services\Data;

use Bazookas\APIFrameworkBundle\Exception\ApiException;
use Bazookas\APIFrameworkBundle\Services\Data\Base\BaseDataService;
use Bazookas\CommonBundle\Exception\NotImplementedException;
use Doctrine\ORM\EntityManager;
use Doctrine\ORM\Query;
use Symfony\Component\HttpFoundation\ParameterBag;
use Symfony\Component\HttpFoundation\Response;
use WebsiteBundle\Entity\Experiment;
use WebsiteBundle\Entity\ExperimentCondition;
use WebsiteBundle\Entity\Participant;
use WebsiteBundle\Entity\Quote;
use WebsiteBundle\Entity\TestAnswer;
use WebsiteBundle\Entity\TestSession;
use WebsiteBundle\Entity\UnlockedQuote;
use WebsiteBundle\Exception\ExceptionMap;

class TestDataService extends BaseDataService
{

  /**
   * @var EntityManager
   */
  private $entityManager;

  private $deviceDetector;

  /**
   * @var SessionDataService
   */
  private $sessionDataService;

  public function __construct(EntityManager $entityManager, $deviceDetector, SessionDataService $sessionDataService) {
    $this->entityManager = $entityManager;
    $this->deviceDetector = $deviceDetector;
    $this->sessionDataService = $sessionDataService;
  }

  /**
   * @param ParameterBag $params
   * @param ParameterBag $extraParams
   * @return array
   */
  public function submitTestResults(ParameterBag $params, ParameterBag $extraParams) {
    /**
     * @var Participant $participant
     */
    $participant = $extraParams->get('user');
    $experiment = $participant->getExperiment();

    if (empty($experiment)) {
      throw new ApiException('This user is not linked to an experiment.', ExceptionMap::SESSION_NO_SESSION, Response::HTTP_FORBIDDEN);
    }

    $data = $params->get('data');
    $session = $this->getSession($participant, $experiment);

    if (empty($session)) {
      throw new ApiException('No session has been scheduled.', ExceptionMap::SESSION_NO_SESSION, Response::HTTP_BAD_REQUEST);
    }

    $now = new \DateTime('now', new \DateTimeZone('UTC'));
    $availableFrom = $session->getAvailableFrom();
    if ($availableFrom > $now) {
      throw new ApiException('Session \''.$session->getNumber().'\' is not yet available. It will be available from ' . $availableFrom->format('Y-m-d H:i:s'), ExceptionMap::SESSION_NOT_YET_AVAILABLE, Response::HTTP_BAD_REQUEST);
    }

    $answers = $this->processAnswers($participant, $session, $data);
    $this->updateSessionStatistics($session, $answers, $params->get('inputType'));

    //Let them unlock a new quote if gamification is enabled
    if ($session->getGamificationLevel() === ExperimentCondition::GAMIFICATION_HIGH) {
      $this->unlockQuote($participant, $experiment, $session);
    }

    //Create the next session
    $minDateTime = new \DateTime('now', new \DateTimeZone('UTC'));
    $minDateTime->modify('+'.$experiment->getMinTimeBetweenSession().' minutes');

    $this->sessionDataService->createNewSession($participant, $experiment, $minDateTime, $session);

    //Check if the user has completed all of their sessions; if so flag them as having completed all of the sessions
    if ($experiment->getTotalSessions() === $session->getNumber()) {
      $participant->setCompletedAllSessions(true);
      $this->entityManager->persist($participant);
      $this->entityManager->flush();
    }

    //Need to return an empty response
    return [];
  }

  /**
   * @param Participant $participant
   * @param Experiment $experiment
   * @return TestSession|null
   */
  private function getSession(Participant $participant, Experiment $experiment) {
    $entityRepo = $this->entityManager->getRepository(TestSession::class);
    return $entityRepo->getNextSession($participant, $experiment, Query::HYDRATE_OBJECT);
  }

  /**
   * @param array $data
   * @param $consecutiveCorrect
   * @param $consecutiveWrong
   * @return TestAnswer
   */
  private function parseAnswerData(array $data, $consecutiveCorrect, $consecutiveWrong) {
    $answer = new TestAnswer();

    $answer
      ->setTrial($data['trial'])
      ->setStimulus($data['stimulus'])
      ->setAnswer($data['answer'])
      ->setCorrectAnswer($data['correctAnswer'])
      ->setCorrect($data['answer'] == $data['correctAnswer'])
      ->setReactionTime($data['reactionTime'])
      ->setInterval($data['interval'])
      ->setLevel($data['level'])
      ->setConsecutiveCorrect($consecutiveCorrect)
      ->setConsecutiveWrong($consecutiveWrong)
      ->setScore($data['score'])
    ;

    return $answer;
  }

  /**
   * @param Participant $participant
   * @param TestSession $session
   * @param array $data
   * @return TestAnswer[]
   */
  private function processAnswers(Participant $participant, TestSession $session, array $data) {
    $BATCH_SIZE = 200;
    $count = 0;
    $consecutiveCorrect = 0;
    $consecutiveWrong = 0;

    $attachedAnswers = [];
    $answers = [];
    foreach($data as $answerData) {
      if ($answerData['answer'] == $answerData['correctAnswer']) {
        $consecutiveCorrect++;
        $consecutiveWrong = 0;
      } else {
        $consecutiveCorrect = 0;
        $consecutiveWrong++;
      }

      $answer = $this->parseAnswerData($answerData, $consecutiveCorrect, $consecutiveWrong);

      $answer
        ->setParticipant($participant)
        ->setSession($session)
      ;

      $this->entityManager->persist($answer);
      $attachedAnswers[] = $answer;
      $answers[] = $answer;

      if ($count % $BATCH_SIZE == 0) {
        $this->entityManager->flush();

        //Detach the objects to free memory
        foreach($attachedAnswers as $key => $item) {
          $this->entityManager->detach($item);
          unset($attachedAnswers[$key]);
        }
      }
    }

    //Persist any items that might not have fit into the last batch
    $this->entityManager->flush();

    return $answers;
  }

  /**
   * @param TestSession $session
   * @param array $answers
   * @param string $inputType
   */
  public function updateSessionStatistics(TestSession $session, array $answers, string $inputType) {
    $minIndex = 0;
    $maxIndex = 0;
    $consecutiveData = [
      'correctTotal' => 0,
      'correctCount' => 0,
      'wrongTotal' => 0,
      'wrongCount' => 0
    ];

    $reactionTime = 0;
    $correctCount = 0;
    $level = 0;
    $totalScore = 0;
    foreach($answers as $index => $answer) {
      if ($answers[$minIndex]->getInterval() > $answer->getInterval()) {
        $minIndex = $index;
      }

      if ($answers[$maxIndex]->getInterval() < $answer->getInterval()) {
        $maxIndex = $index;
      }

      if ($answers[$index]->getLevel() > $level) {
        $level = $answers[$index]->getLevel();
      }

      $consecutiveData = $this->calculateConsecutiveData($index, $answers, $consecutiveData);

      $reactionTime = $reactionTime + $answer->getReactionTime();

      $totalScore = $totalScore + $answer->getScore();

      if ($answer->isCorrect()) {
        $correctCount = $correctCount + 1;
      }
    }

    //add the last consecutive values
    $value = $answers[$index]->getConsecutiveCorrect();
    if ($value > 0) {
      $consecutiveData['correctTotal'] = $consecutiveData['correctTotal'] + $value;
      $consecutiveData['correctCount']++;
    }

    $value = $answers[$index]->getConsecutiveWrong();
    if ($value > 0) {
      $consecutiveData['wrongTotal'] = $consecutiveData['wrongTotal'] + $value;
      $consecutiveData['wrongCount']++;
    }

    //find the min and max intervals
    $min = $answers[$minIndex]->getInterval();
    $max = $answers[$maxIndex]->getInterval();

    //Calculate the median of the intervals
    usort($answers, [$this, 'compareInterval']);

    $answerCount = count($answers);
    $medianIndex = $answerCount / 2;
    if ($answerCount % 2 ==0 ) {
      //its even
      $median = ($answers[$medianIndex]->getInterval() + $answers[$medianIndex - 1]->getInterval()) / 2;
    } else {
      //its odd
      $median = $answers[$medianIndex]->getInterval();
    }

    $avgConsecutiveCorrect = ($consecutiveData['correctTotal'] / ($consecutiveData['correctCount'] > 0 ? $consecutiveData['correctCount'] : 1));
    $avgConsecutiveWrong = ($consecutiveData['wrongTotal'] / ($consecutiveData['wrongCount'] > 0 ? $consecutiveData['wrongCount'] : 1) );

    $session
      ->setMinInterval($min)
      ->setMaxInterval($max)
      ->setMedianInterval($median)
      ->setAvgConsecutiveCorrect($avgConsecutiveCorrect)
      ->setAvgConsecutiveWrong($avgConsecutiveWrong)
      ->setCompleted(new \DateTime('now', new \DateTimeZone('UTC')))
      ->setTrialCount($answerCount)
      ->setCorrectCount($correctCount)
      ->setMeanReactionTime(($reactionTime/ $answerCount))
      ->setDevice($this->deviceDetector->isTablet() ? TestSession::DEVICE_TABLET : TestSession::DEVICE_PC)
      ->setInputType($inputType)
      ->setFinalLevel($level)
      ->setTotalScore($totalScore)
    ;

    $this->entityManager->persist($session);
    $this->entityManager->flush();
  }

  /**
   * @param int $index
   * @param TestAnswer[] $answers
   * @param array $consecutiveData
   * @return array
   */
  private function calculateConsecutiveData($index, array $answers, array $consecutiveData) {
    $answer = $answers[$index];
    if ($index > 0 && $answer->getConsecutiveCorrect() === 0) {
      $value = $answers[$index -1]->getConsecutiveCorrect();
      if ($value > 0) {
        $consecutiveData['correctTotal'] = $consecutiveData['correctTotal'] + $value;
        $consecutiveData['correctCount']++;
      }
    }

    if ($index > 0 && $answer->getConsecutiveWrong() === 0) {
      $value = $answers[$index -1]->getConsecutiveWrong();
      if ($value > 0) {
        $consecutiveData['wrongTotal'] = $consecutiveData['wrongTotal'] + $value;
        $consecutiveData['wrongCount']++;
      }
    }

    return $consecutiveData;
  }

  /**
   * Compare the interval between two Test answers
   * @param TestAnswer $a
   * @param TestAnswer $b
   * @return int
   */
  private function compareInterval(TestAnswer $a, TestAnswer $b) {
    if ($a->getInterval() > $b->getInterval()) {
      return 1;
    } else if ($a->getInterval() < $b->getInterval()) {
      return -1;
    } else {
      return 0;
    }
  }

  /**
   * Assign a new unlocked quote to the user
   * @param Participant $participant
   * @param Experiment $experiment
   * @param TestSession $session
   */
  private function unlockQuote(Participant $participant, Experiment $experiment, TestSession $session) {
    $unlockedQuotes = $participant->getUnlockedQuotes();
    $unlocked = [];
    foreach($unlockedQuotes as $unlockedQuote) {
      $unlocked[] = $unlockedQuote->getQuote()->getId();
    }

    $experimentRepo = $this->entityManager->getRepository(Experiment::class);

    //Check if it's the last session, if it is assign the master quote
    $quoteType = Quote::TYPE_GENERIC;
    if ($session->getNumber() === $experiment->getTotalSessions()) {
      $quoteType = Quote::TYPE_MASTER;
    }

    $remainingQuotes = $experimentRepo->getRemainingQuotes($unlocked, $experiment, $quoteType);

    if (!empty($remainingQuotes)) {
      $index = rand(0, count($remainingQuotes) - 1);
      $newQuote = new UnlockedQuote();
      $newQuote
        ->setParticipant($participant)
        ->setQuote($remainingQuotes[$index])
        ->setSession($session)
      ;

      $this->entityManager->persist($newQuote);
      $this->entityManager->flush();
    }
  }
}
