<?php

namespace Bazookas\AdminBundle\Controller\Base;

use Bazookas\AdminBundle\PageBuilder\Interfaces\FormPageBuilderInterface;
use Bazookas\AdminBundle\PageBuilder\Interfaces\PageBuilderInterface;
use Bazookas\CommonBundle\Entity\Interfaces\EntityInterface;
use Bazookas\CommonBundle\Exception\Interfaces\EntityExceptionInterface;
use Bazookas\CommonBundle\Utility\Utility;
use Symfony\Component\DependencyInjection\ContainerAwareInterface;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpKernel\Exception\AccessDeniedHttpException;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use Symfony\Component\Security\Core\Exception\AccessDeniedException;

abstract class BaseAdminActionController extends BaseAdminController
{
  //Page action constants
  const ACTION_DEFAULT = self::ACTION_LIST;
  const ACTION_LIST = 'list';
  const ACTION_EDIT = 'edit';
  const ACTION_ADD = 'add';
  const ACTION_VIEW = 'view';
  const ACTION_REMOVE = 'remove';
  const ACTION_BULK_EDIT = 'bulkEdit';
  const ACTION_BULK_REMOVE = 'bulkRemove';
  const ACTION_EXPORT = 'export';
  const ACTION_DASHBOARD = 'dashboard';

  //Ajax action constants
  const AJAX_LIST = 'ajaxList';
  const AJAX_INLINE_EDIT = 'ajaxInlineEdit';

  /**
   * @var PageBuilderInterface[]
   */
  protected $builders;

  /**
   * This function is the entry point for most pages, it will attempt to route the
   * request to a custom action function and failing that handle it in a default manner.
   * @param Request $request
   * @param null $_action
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   */
  public function indexAction(Request $request, $_action = null)
  {
    //If the action is empty set it to the default action. This constant can be overwritten in the controller as needed.
    if (empty($_action)) {
      //Using a static reference allows child classes to overwrite the default action constant
      $_action = static::ACTION_DEFAULT;
    }

    // before moving on check if the current user has access
    if (!$this->hasAccess($_action)) {
      return $this->accessDeniedRedirect($request, $request->get('_route'), [
        '_action' => static::ACTION_DEFAULT
      ]);
    }

    //get the action function that should be carried out
    $actionFunction = $this->getActionFunction($request, $_action);

    //attempt to execute the action function
    try {
      return $this->$actionFunction($request, $_action);
    } catch(AccessDeniedHttpException $e) {
      return $this->accessDeniedRedirect($request, $request->get('_route'), [
        '_action' => static::ACTION_DEFAULT
      ]);
    } catch(EntityExceptionInterface $e) {
      return $this->entityAccessDeniedRedirect($request, $e, $request->get('_route'), [
        '_action' => static::ACTION_DEFAULT
      ]);
    }
  }

  /**
   * This function returns a string containing the function name of the action to be invoked
   * @param Request $request
   * @param string $action
   * @return string
   */
  protected function getActionFunction(Request $request, $action) {
    $actionFunction = $action . 'Action';
    if (method_exists($this, $actionFunction)) {
      return $actionFunction;
    } else {
      return 'defaultAction';
    }
  }

  /**
   * A default action which is able to render a standard page
   * @param Request $request
   * @param string $action
   * @return \Symfony\Component\HttpFoundation\Response
   */
  protected function defaultAction(Request $request, $action) {
    //Get the page builder for the current action
    $builder = $this->getPageBuilder($action);

    //Check if a page builder was defined
    if (empty($builder)) {
      throw new NotFoundHttpException();
    }

    if ($builder instanceof PageBuilderInterface) {
      //Allow the builder to be modified using the modify hook
      $builder = $this->modifyBuilder($request, $action, $builder);

      //Build the page and add its built elements to the current page.
      $builder->buildPage($request);
      $this->elements = $builder->getElements();

      //TODO Add post build hook?

      //Get any custom template data
      $data = $builder->getTemplateData();
    } else if (is_callable($builder)){
      $data = $builder();
    } else {
      throw new NotFoundHttpException();
    }

    //Are we dealing with a picker item?
    if ($request->get('picker')) {
      $this->layoutTemplate = self::TEMPLATE_LAYOUT_MENU_LESS;
    }

    //Merge the custom template data with the default template data
    $templateData  = $this->getTemplateData($request, $action, $data);

    //Generate the response
    return $this->createResponse($request, $templateData, $action);
  }

  /**
   * Find the page builder for the current request based on the action.
   * @param string $action
   * @return PageBuilderInterface|callable|null
   * @throws \Exception
   */
  protected function getPageBuilder($action) {
    if (isset($this->builders[$action])) {
      $builder = $this->builders[$action];

      //Allow service definitions to be used as well, these must be strings that start with @
      if (Utility::startsWith($builder, '@')) {
        //Retrieve the builder
        $builder = $this->get(substr($builder, 1, strlen($builder)));

        if ($builder instanceof PageBuilderInterface) {
          $entityManager = $this->get('doctrine.orm.entity_manager');

          //Set the default page builder properties
          $builder
            ->setEntityManager($entityManager)
            ->setEntityClass($this->getEntityClass())
            ->setElements($this->elements)
          ;

          if ($builder instanceof ContainerAwareInterface) {
            $builder->setContainer($this->container);
          }

          return $builder;
        } else {
          //Throw an error as the service is improperly configured
          throw new \Exception('The defined page builder service must implement \'' . PageBuilderInterface::class . '\'');
        }
      }

      //Check if its a page builder fully qualified class name
      if (in_array(PageBuilderInterface::class, class_implements($builder))) {
        $entityManager = $this->get('doctrine.orm.entity_manager');

        $builder = new $builder();
        $builder
          ->setEntityManager($entityManager)
          ->setEntityClass($this->getEntityClass())
          ->setElements($this->elements)
        ;

        if ($builder instanceof ContainerAwareInterface) {
          $builder->setContainer($this->container);
        }
      }

      return $builder;
    }

    return null;
  }

  /**
   * Modify the current page builder if a custom modification function exists. The function must always start
   * with 'modify' and end with 'Builder'; eg. 'modifyListBuilder'
   * @param Request $request
   * @param $action
   * @param PageBuilderInterface $builder
   * @return PageBuilderInterface
   */
  protected function modifyBuilder(Request $request, $action, PageBuilderInterface $builder) {
    $modifyFunction = 'modify' . ucfirst($action) . 'Builder';

    if (method_exists($this, $modifyFunction)) {
      return $this->$modifyFunction($request, $builder);
    } else {
      return $builder;
    }
  }

  /**
   * A standard function for the handling of a form
   * @param Request $request
   * @param $action
   * @return \Symfony\Component\HttpFoundation\Response
   * @throws \Exception
   */
  protected function formAction(Request $request, $action) {
    //Retrieve the page builder
    $builder = $this->getPageBuilder($action);

    //Allow the builder to be modified
    $builder = $this->modifyBuilder($request, $action, $builder);

    if (!$builder instanceof FormPageBuilderInterface) {
      throw new \Exception('The form builder must implement the interface: ' . FormPageBuilderInterface::class);
    }

    //Handle the form
    $success = $builder->handleForm($request);
    if ($success) {
      //Redirect and add a flash message
      $translator = $this->get('translator');
      $messageKey = $this->getActionFlashMessageKey($action, $builder);
      $this->addFlash('success', $translator->trans($messageKey, [], 'admin'));

      $picker = $request->get('picker');

      if ($picker) {
        return $this->redirectToRoute($request->get('_route'), [
          'picker' => $request->get('picker')
        ]);
      } else {
        return $this->redirectToRoute($request->get('_route'));
      }
    }

    //Build the page
    $builder->buildPage($request);

    //Add the built elements to the current page
    $this->elements = $builder->getElements();

    //Get any custom template data
    $data = $builder->getTemplateData();

    //Merge the custom template data with the default template data
    $templateData  = $this->getTemplateData($request, $action, $data);

    //Generate a response
    return $this->createResponse($request, $templateData, $action);
  }

  /**
   * @param string $action
   * @param string|PageBuilderInterface $entityName
   * @return string the translation key
   */
  protected function getActionFlashMessageKey($action, $entityName) {
    switch($action) {
      case self::ACTION_ADD:
        if ($entityName instanceof PageBuilderInterface) {
          return 'admin.entities.'. lcfirst($entityName->getEntityShortClassName()) . '.onAdd';
        } else {
          return 'admin.entities.'. $entityName . '.onAdd';
        }
      case self::ACTION_EDIT:
      case self::ACTION_BULK_EDIT:
        if ($entityName instanceof PageBuilderInterface) {
          return 'admin.entities.'. lcfirst($entityName->getEntityShortClassName()) . '.onEdit';
        } else {
          return 'admin.entities.'. $entityName . '.onEdit';
        }
      case self::ACTION_REMOVE:
        return 'admin.entities.'. $entityName . '.onRemoved';
      default:
        return '';
    }
  }

  /**
   * @param Request $request
   * @param EntityExceptionInterface $e
   * @param string|null $route
   * @param array|null $routeParams
   * @return \Symfony\Component\HttpFoundation\RedirectResponse
   */
  public function entityAccessDeniedRedirect(Request $request, EntityExceptionInterface $e, $route = null, $routeParams = null) {
    $translator = $this->get('translator');

    $message = $translator->trans('admin.list.messages.entityAccessDenied', [], 'admin');
    $payload = $e->getPayload();
    if (is_array($payload)) {
      $errors = [];
      foreach($e->getPayload() as $entityName => $items) {
        foreach($items as $item) {
          if ($item instanceof EntityInterface) {
            $errors[$message][] = $item->toHumanReadable($request->getLocale());
          }
        }
      }
    } else if ($payload instanceof EntityInterface) {
      $errors[$message][] = $payload->toHumanReadable($request->getLocale());
    }

    if (empty($errors)) {
      $errors = $message;
    }

    $this->addFlash(
      'error', $errors
    );

    //send the user back where they belong!
    if (empty($route)) {
      $route = $request->get('_route');
    }
    if (is_null($routeParams)) {
      $routeParams = $request->get('_route_params');
    }

    if (empty($route)) {
      throw new AccessDeniedException();
    }

    return $this->redirectToRoute($route, $routeParams);
  }

  /**
   * @return string the entity fully qualified class name
   */
  protected abstract function getEntityClass();

}
