## General information
### Controller architecture

The admin bundle provides a number of base admin controllers. The controllers you will be working with directly will most often be the `BaseAdminListController` or the `BaseAdminDashboardController`. These classes provide some convenience functionality for creating a basic list table or dashboard respectively. Their core functionality however comes from the `BaseAdminActionController`.

The `BaseAdminActionController` is responsible for processing an action through the `indexAction` function. This will always be the entry point for all routes in your CMS system. Before expanding on the functionality of the `BaseAdminActionController` we should first take a closer look at the relationship between all of the abstract admin controllers.

![](./Images/controller_architecture.jpg)

As you can see the `BaseAdminController` is the most abstract class containing the least amount of functionality. In most cases you will therefore be using the convenience classes `BaseAdminListController` or `BaseAdminDashboardController`. So what does the action controller do?

The action controller extends the basic functionality provided in `BaseAdminController` by adding an action framework. In essence this provides an `indexAction` function which will serves as the entry point for all admin actions linked to that controller. The only action it defines itself is `ACTION_DEFAULT`. This constant can be overwritten in your child controller to set the default action. By simply re-defining the constant:

```php
const ACTION_DEFAULT = 'list';
```

### The index action

The index action will attempt to find a an action function to execute for the current action, if it cannot find a custom action function it performs the `defaultAction` function. For example let us assume that the we are attempting to view a list table page. The action in this case would be the constant `ACTION_LIST` with the value `list`. The `indexAction` will therefore attempt to execute a function called `listAction` or failing to find this custom function, will execute the `defaultAction` function. 

The main difference here is that a custom action function allows you to create completely custom pages as it function in the exact same way as normal Symfony controller action. On the other hand the default action allows you to make use of the Page Builder system ([more info below](#thepagebuildersystem)) to render the page. The advantage of this is that the page builder system allows you to rapidly create fairly standard pages using a minimum of custom code. You should therefore create a custom action function if you need to do something highly customised and use the page builder when using more standard functionality. Note that it is always possible to mix the two methods.

### Custom action functions

A custom action function can be defined in the child controller in order to create custom pages. You define a custom action by taking the `ACTION_CONSTANT` + 'Action' as your function name. For example:

```php
<?php

use Symfony\Component\HttpFoundation\Request;
use Bazookas\AdminBundle\Controller\Base\BaseAdminActionController;

class MyController extends BaseAdminActionController {
  
  //Let's say you've defined a constant
  const ACTION_LIST = 'list';
  const ACTION_OVERVIEW = 'overview';
  
  //To perform custom functionality for the list action the function can be overwritten by creating your own function
  public function listAction(Request $request, $action) {
    //... 
  }
  
  protected function getEntityClass()
  {
    return EntityClass::class;
  }
    
  protected function hasAccess($action) {
    return true;
  }
}
````

Please note that this is applicable to all types of actions including custom actions. Meaning that adding new custom actions in your controller is completely supported by both the custom action functionality and the page builder system. An example of this is the `ACTION_OVERVIEW` constant we've defined in the example above. This action is completely compatible with the page builder and the custom action functionality. In this case the custom action would be defined by creating a function called `overviewAction`.

### The page builder system

When not defining your own actions you can make use of the page builder system. The page builder system allows you to define page builder for each of your actions. The page builders are classes, or services which are responsible for creating the AdminElements that will be rendered on your page. 

The admin bundle provides several pre-defined page builders:
- `ListPageBuilder`
- `DashboardPageBuilder`
- `BaseFormPageBuilder`

For brevity we will only discuss the three most commonly used page builders but do note that other page builders exist.
