## Admin list controller

The `BaseAdminListController` is the class you will be working with the most and which you will be extending in all of your project specific controllers when creating list views. It performs most of the functionality the old `ListBaseController` performed.

### Defining a route

Let's start by defining a route for our new controller:
```yaml
admin_test:
  path:     /test/{_action}/{_id}
  defaults:
    _controller: WebsiteBundle:Admin/TestAdmin:index
    _action: ~
    _id: ~
```

Your path can be anything as long as it contains the `{_action}` and `{_id}` parameters. Also note that the `indexAction` will almost always be the entry point to your controller for any action. Now that we have a route we must defined our own, project specific, controller.

### Defining a controller

In order create our own controller we must implement 4 things defined in the abstract class `BaseAdminListController`:

1. Define which entity this list view will use by implementing the `getEntityClass` function
2. Define which actions are available by implementing the `hasAccess` function
3. Modify the list builder to add the fields that should be included in the list view table.
4. Define which form classes to use for which action (more on this later)

Below is an example of a basic implementation of the AdminListController:

```php
<?php

namespace WebsiteBundle\Controller\Admin;

use Bazookas\AdminBundle\Controller\Base\BaseAdminListController;
use Bazookas\AdminBundle\PageBuilder\ListPageBuilder;
use Symfony\Component\HttpFoundation\Request;
use WebsiteBundle\Entity\Test;

class TestAdminController extends BaseAdminListController
{

  protected function getEntityClass()
  {
    return Test::class;
  }

  protected function modifyListBuilder(Request $request, ListPageBuilder $builder)
  {
    $builder = parent::modifyListBuilder($request, $builder);

    $builder
      ->addField('name')
      ->addField('type', [
        'template' => 'WebsiteBundle:List\Fields:testType.html.twig'
      ])
    ;
    
    //Pass the user to the list query builder
    $builder->setListQueryParams([
      'user' => $this->getUser()
    ]);
    
    return $builder;
  }

  protected function getFormClass($action) {
    //Leave this for now
    return null;
  }

  protected function hasAccess($action) {
    switch($action) {
      case self::ACTION_LIST:
        return true;
      default:
        return false;
    }
  }
}
```

#### Setting the entity class

Let's start by looking at the `getEntityClass` function. This function declares the entity which will be used to render the list view. This is important to note as the underlying base controllers use this to retrieve an EntityRepository for the defined class. Note that the EntityRepository class must always implement the [ListRepositoryInterface](Repository\Interfaces\ListRepositoryInterface) a basic implementation of this interface is provided by the `BaseRepository` class.

The two most important things which the interface requires are:

```php
  /**
   * @param array $fields
   * @param array $params
   * @return QueryBuilder
   */
  public function createListQueryBuilder(array $fields, array $params);

  /**
   * @param QueryBuilder $qb
   * @param array $params
   * @return QueryBuilder
   */
  public function beforeListQueryExecution(QueryBuilder $qb, array $params);
```

You can overwrite the base implementation of these functions in your own repository in order to customise what data is used in the list table. Note that the parameters `params` passed to the function can be customised through the builder modification process by setting them through `$builder->setListQueryParams($items);`. An example is passing the currently logged in user to the query builder (as shown in the above example).

#### Page builder modification
The page builder is responsible for creating the admin elements that should be added to the page. A page builder can be defined for an action by adding it to builders array in the constructor of the controller:

```php
  public function __construct() {
    parent::__construct();

    $this->builders[BaseAdminListController::ACTION_LIST] = ListPageBuilder::class;
  }
```

The builders array is an associative array consisting of `<action> => <builder>` elements. Note that a page builder can be defined using a fully qualified class name (as shown above) or it can be defined as a callable or as a service. To declare a page builder as a service provide the service id as a string preceded by an @ (for example "@bazookas.services.myService"). Note that the service must be an instance of `PageBuilderInterface`. For a more in-depth view of page builder see the `Core concepts` section.

In the case of a `ListPageBuilder` the page builder modification process allows you to defined which fields are shown in the list table. 

In the earlier example we defined 2 columns:

```php
  $builder
    ->addField('name')
    ->addField('type', [
      'template' => 'WebsiteBundle:List\Fields:testType.html.twig'
    ])
  ;
```

The builder automatically adds a label for the field in the conventional format: `admin.entities.<my entity>.fields.<field name>`. As shown in the example above the template for each `ListFieldElement` can be defined. For more options see the `ListFieldElement`.

### Add or edit functionality

Now that we've set up a basic controller let's go ahead and add functionality for adding or editing an entity.

In order for the add or edit actions to work we need to first enable them in the `hasAccess` function:

```php
  protected function hasAccess($action) {
    switch($action) {
      case self::ACTION_LIST:
      case self::ACTION_EDIT:
      case self::ACTION_ADD:
        return true;
      default:
        return false;
    }
  }
```
 Then we need to define which class should be used to render the form. We do this in the `getFormClass` function:
 
 ```php
  protected function getFormClass($action) {
     return MyEntityAdminType::class;
   }
 ```
 
 Congratulations! This all you need to do to get the add and edit functionality to work.
 
 ### Enabling remove functionality
 
 The remove functionality can be enabled through the `hasAccess` function:
 
 ```php
   protected function hasAccess($action) {
     switch($action) {
       case self::ACTION_LIST:
       case self::ACTION_EDIT:
       case self::ACTION_ADD:
       case self::ACTION_REMOVE:
         return true;
       default:
         return false;
     }
   }
 ```

And that's all that's needed to enable the remove functionality.
