<?php

namespace Bazookas\APIFrameworkBundle\Formatter\Base;
use Bazookas\APIFrameworkBundle\Exception\FormatterException;
use Bazookas\CommonBundle\Utility\Utility;

class OutputFormatter
{
  //conversion types
  const TYPE_STRING = 'string';
  const TYPE_INTEGER = 'integer';

  /**
   * @var array
   */
  protected $config;

  /**
   * @var array
   */
  protected $globalConfig;

  public function __construct(array $config, array $globalConfig = []) {
    $this->config = $config;
    $this->globalConfig = $globalConfig;
  }

  /**
   * @param array $data
   * @return array
   */
  public function format(array &$data) {
    if (empty($data)) {
      return $data;
    }

    if (!Utility::isAssociative($data)) {
      //If we're dealing with a numerically indexed array loop over the items and apply the field transformation to each item
      foreach($data as &$item) {
        //Carry out field transformations
        $item = $this->applyFieldTransformations($item, $this->config);

        //Carry out global transformations
        $item = $this->applyGlobalTransformations($item, $this->globalConfig);
      }
    } else {
      //Carry out field transformations
      $data = $this->applyFieldTransformations($data, $this->config);

      //Carry out global transformations
      $data = $this->applyGlobalTransformations($data, $this->globalConfig);
    }

    return $data;
  }

  /**
   * @param array $data
   * @param array $configuration
   * @return array
   */
  public function applyFieldTransformations(array $data, array $configuration) {
    if (empty($configuration)) {
      return $data;
    }

    foreach($configuration as $field => $transformationConfig) {
      try {
        //Are we dealing with a transformation object or an array of transformation objects
        if (Utility::isAssociative($transformationConfig)) {

          //Is the field we want to apply the transformation on defined? and are we dealing with an associative array or a numerically indexed array
          if (!isset($data[$field])) {
            continue;
          }

          $data[$field] = $this->executeTransformation($data[$field], $transformationConfig);
        } else {
          foreach ($transformationConfig as $config) {
            $data = $this->applyFieldTransformations($data, [$field => $config]);
          }
        }
      } catch(FormatterException $e) {
        throw new FormatterException('Transformation error for field \''.$field.'\': ' . $e->getMessage());
      }
    }

    return $data;
  }

  /**
   * @param array $data
   * @param array $configuration
   * @return array
   */
  public function applyGlobalTransformations(array $data, array $configuration) {
    if (empty($configuration)) {
      return $data;
    }

    foreach($configuration as $transformationConfig) {
      if (Utility::isAssociative($transformationConfig)) {
        try {
          $data = $this->executeTransformation($data, $transformationConfig, true);
        } catch (FormatterException $e) {
          throw new FormatterException('Global transformation error: ' . print_r($e->getMessage(), true));
        }
      } else {
        $data = $this->applyGlobalTransformations($data, $transformationConfig);
      }
    }

    return $data;
  }

  /**
   * @param mixed $data
   * @param array $configuration
   * @param bool $isGlobal
   * @return array
   */
  private function executeTransformation($data, array $configuration, $isGlobal = false) {
    $transformation = isset($configuration['transformation']) ? $configuration['transformation'] : null;

    //Is a transformation defined?
    if (empty($transformation)) {
      throw new FormatterException('No transformation defined');
    }

    $params = isset($configuration['params']) ? $configuration['params'] : [];
    if (!is_array($params)) {
      $params = [$params];
    }

    if (!is_array($data) || Utility::isAssociative($data) || $isGlobal) {
      return $this->transformData($data, $params, $transformation);
    } else {
      foreach($data as $key => &$item) {
        $data[$key] = $this->transformData($item, $params, $transformation);
      }

      return $data;
    }
  }

  /**
   * @param $data
   * @param $params
   * @param $transformation
   * @return array|mixed
   */
  private function transformData($data, $params, $transformation) {
    array_unshift($params, $data);

    if (is_callable($transformation)) {
      return call_user_func_array($transformation, $params);
    } else if ($transformation instanceof OutputFormatter) {
      return $transformation->format($data);
    } else {
      throw new FormatterException('The transformation must be a callable or an instance of \''. OutputFormatter::class .'\'.');
    }
  }

  /* ---------------------------------------------------------------------------------------------
   *                              Default field transformation functions
   * ---------------------------------------------------------------------------------------------- */
  /**
   * @param $value
   * @param string $type
   * @return int|string
   */
  protected function convertToType($value, string $type) {
    switch ($type) {
      case self::TYPE_STRING:
        return (string) $value;
      case self::TYPE_INTEGER:
        return (int) $value;
      default:
        return $value;
    }
  }

  /**
   * @deprecated to be removed in v2
   */
  protected function propertyToScalarArray($data, $property = 'id') {
    return $this->propertyToScalar($data, $property);
  }

  /**
   * @param array $element
   * @param string $property
   * @return mixed
   */
  protected function propertyToScalar(array $element, string $property = 'id') {
    return isset($element[$property]) ? $element[$property] : null;
  }

  /* ---------------------------------------------------------------------------------------------
   *                              Default global transformation functions
   * ---------------------------------------------------------------------------------------------- */

  /**
   * @param array $data
   * @param string $field
   * @return mixed
   * @throws \Exception
   */
  protected function addToParent(array $data, string $field) {
    if (!isset($data[$field])) {
      return $data;
    }

    foreach($data[$field] as $key => $item) {
      if (!isset($data[$key])) {
        $data[$key] = $item;
      } else {
        throw new \Exception('The key \''.$key.'\' already exists in \''.print_r($data, true).'\'');
      }
    }

    //remove the previous parent key
    unset($data[$field]);

    return $data;
  }

  /**
   * @param array $data
   * @param string $field
   * @return mixed
   */
  protected function convertArrayToSingleton(array $data, string $field) {
    if (isset($data[$field])) {
      $data[$field] = $data[$field][0];
    }

    return $data;
  }

  /**
   * @param array $data
   * @param string[] $fields
   * @return mixed
   */
  protected function unsetFields(array $data, array $fields) {
    if (!is_array($fields)) {
      $fields = [$fields];
    }

    foreach($fields as $field) {
      unset($data[$field]);
    }

    return $data;
  }

  /**
   * @param array $data
   * @param string[] $fields
   * @return mixed
   * @throws \Exception
   */
  protected function replaceKeyWithValue(array $data, array $fields) {
    if (!Utility::isAssociative($fields)) {
      throw new \Exception('Excepted association array \'[<oldKey> => <newKey>]\' got \''.print_r($fields, true).'\'');
    }

    foreach($fields as $oldKey => $newKeyIndex) {
      if (isset($data[$oldKey])) {
        $newKey = $data[$oldKey][$newKeyIndex];
        $data[$newKey] = $data[$oldKey];
        unset($data[$oldKey]);
      }
    }

    return $data;
  }

  /**
   * @param array $data
   * @param callable $param
   * @return mixed
   */
  protected function transformFieldKey(array $data, callable $param) {
    foreach($data as $key => $value) {
      if (is_callable($param)) {
        $newKey = call_user_func($param, $key);
        $data[$newKey] = $value;
        unset($data[$key]);
      }
    }

    return $data;
  }

  /**
   * @param array $data
   * @param string[] $fields
   * @return mixed
   * @throws \Exception
   */
  protected function renameKey(array $data, array $fields) {
    if (!Utility::isAssociative($fields)) {
      throw new \Exception('Excepted association array \'[<oldKey> => <newKey>]\' got \''.print_r($fields, true).'\'');
    }

    foreach($fields as $oldKey => $newKey) {
      if (isset($data[$oldKey])) {
        $data[$newKey] = $data[$oldKey];
        unset($data[$oldKey]);
      }
    }

    return $data;
  }

  /**
   * @param array $data
   * @param string $field
   * @return array
   */
  protected function usePropertyValueAsKey(array $data, string $field) {
    if (empty($data)) {
      return $data;
    }

    if (Utility::isAssociative($data)) {
      $value = $data[$field];
      unset($data[$field]);
      return [$value => $data];
    } else {
      foreach($data as $key => $item) {
        $value = $item[$field];
        unset($item[$field]);
        $data[$value] = $item;
        unset($data[$key]);
      }

      return $data;
    }
  }

  /**
   * Adds an array of translated objects to the parent entity as $parent[field] = [ $locale => value ] pairs.
   * @param array $data
   * @param string $field
   * @return array
   * @throws \Exception
   */
  public function addTranslationsToParent(array $data, string $field) {
    if (!Utility::isAssociative($data[$field])) {
      foreach($data[$field] as $fieldItem) {
        if (!isset($fieldItem['locale'])) {
          continue;
        }

        $locale = $fieldItem['locale'];
        foreach($fieldItem as $key => $value) {
          if ($key === 'locale') {
            continue;
          }

          $data[$key][$locale] = $value;
        }
      }
    } else {
      throw new \Exception('The function addTranslationsToParent can only be used on numerically indexed arrays');
    }

    unset($data[$field]);
    return $data;
  }
}
