## Formatting output

### How does it work

Its often the case that the output from your repository calls does not match the expected output of the api call. Often you'll need to remove or condense fields in order to make the json output more readable and easier to parse. This component is designed to provide easy to use formatting of your api output.

Let us start with the basics. A formatter can be created in two ways: inline when creating a new object or as a separate class. When performing a lot or complex formatting tasks its preferred that the formatter be moved to its own class and subdivided if necessary.

Regardless of whether or not you're creating an inline object or a separate class the [Bazookas\APIFrameworkBundle\Formatter\Base\OutputFormatter])(../../Formatter/Base/OutputFormatter.php) is the basis for all your formatter objects. Every formatter consists of two components: a `config` array and a `globalConfig` array.

```php
new OutputFormatter($config, $globalConfig);
```

So lets assume we have the following data:

```
$miniGame = [
    'id' => 1,
    'difficulty' => 3,
    'details' => [
        [
            'id' => 1,
            'name' => 'test
        ]
    ],
    'created' => new \DateTime(),
    'modified' => new \DateTime()
];
```

In this example we have a minigame which has its own fields and an associated details entity. Now the details value is an array of objects, but let us assume we have filtered the data by locale and we'll always only have 1 details object that will be returned. It's now pointless to send an array of objects if we're only expecting a single object every time. 

We can condense the output by combining the properties of the details object and its parent object. We do this by defining a global config as follows:

```php
$globalConfig = 
  array(
      array(
        'transformation' => [$this, 'unsetFields'],
        'params' => [['modified', 'created']]
      ),
      array(
        'transformation' => [$this, 'addToParent'],
        'params' => ['details']
      )
  );
```

In the above configuration we've told the formatter to apply a `global` transformation. This means the transformation will be applied to the global scope of the formatter in this example that would be the entire miniGame object. Transformations are always executed in index order. This means that in the example above the `unsetFields` function will be executed before the `addToParent` function. 

Each transformation consist of an array object with a `transformation` key and a `params` key. `conversionFormat` should contain a php callable, that is any function that can be executed from php itself or any object function as shown in the example above. The first transformation in the example will in essence execute a OutputFormatter function called `unsetFields` and give to that function the data its formatting as well as any parameters defined in `params`. In the example above this will cause the `created` and the `modified` fields to be removed from the miniGame object.

The second transformation will add the contents of the details field to the parent object and will remove the details field itself. After applying the above transformation using the following code:

```php
$miniGame = [
    'id' => 1,
    'difficulty' => 3,
    'details' => [
        [
            'name' => 'test'
        ]
    ],
    'created' => new \DateTime(),
    'modified' => new \DateTime()
];

$globalConfig = 
  array(
      array(
        'transformation' => [$this, 'unsetFields'],
        'params' => [['modified']]
      ),
      array(
        'transformation' => [$this, 'addToParent'],
        'params' => ['details']
      )
  );
  
  $formatter = new OutputFormatter($config, $globalConfig);
  $formatter->format($miniGame);
```

This will change the minigame object to:
 ```php
 $miniGame = [
     'id' => 1,
     'difficulty' => 3,
     'name' => 'test,
     'created' => new \DateTime()
 ];
 ```
 and will, when json_encoded, change the output of that array from:
 
 ```json
{
   "id":1,
   "difficulty":3,
   "details":[
      {
         "name":"test"
      }
   ],
   "created":{
      "date":"2016-12-06 19:07:58.000000",
      "timezone_type":3,
      "timezone":"Europe\/Zurich"
   }
}
 ```

to

 ```json
  {
   "id": 1,
   "difficulty": 3,
   "name": "test",
   "created":{
         "date":"2016-12-06 19:07:58.000000",
         "timezone_type":3,
         "timezone":"Europe\/Zurich"
  }
  }
 ```

 
Be aware that in the example above the details entity did not contain an `id` field, in most situations this will however be the case. This is important to note because the `addToParent` function will throw an error if the parent already contains the a property with the same key as the child. This means that in a situation where you have:
 
 ```json
 {
    "id":1,
    "difficulty":3,
    "details":[
       {
          "id":1,
          "name":"test"
       }
    ],
    "created":{
       "date":"2016-12-06 19:07:58.000000",
       "timezone_type":3,
       "timezone":"Europe\/Zurich"
    }
 }
 ```
 
 You will first need to unset the id before you can perform the `addToParent` operation.

 Using the `config` array we can apply field specific transformations. Note that all field specific transformations are always executed before global transformations. Lets assume we wanted to convert the `created` field datetime object to a unix timestamp. In order to do this we define the config as follows and then apply the formatting.
 
 ```php
$config = [
  'created' => [
          [
            'transformation' => 'date_format',
            'params' => 'U'
          ],
          [
            'transformation' => [OutputFormatter::class, 'convertToType'],
            'params' => OutputFormatter::TYPE_INTEGER
          ]
        ]
];

$globalConfig = 
  array(
      array(
        'transformation' => [$this, 'unsetFields'],
        'params' => [['modified']]
      ),
      array(
        'transformation' => [$this, 'addToParent'],
        'params' => ['details']
      )
  );
  
$formatter = new OutputFormatter($config, $globalConfig);
$formatter->format($miniGame);
 ```
 
 When the resulting array is converted to json the output of the above formatter is:
 
 ```json
 {
    "id":1,
    "difficulty":3,
    "name": "test",
    "created":1481047820
 }
 ```
 
### Defining a formatter as a separate class
 A formatter class is a class that extends the the base formatter. The `ChallengerFormatter` in the example below will therefore look similar to an inline formatter definition as it to uses the `config` and `globalConfig` parameters to define its function.

```php
<?php

namespace APIBundle\Formatter\Challenger;

use Bazookas\APIFrameworkBundle\Formatter\Base\OutputFormatter;

class ChallengerFormatter extends OutputFormatter
{
  
  public function __construct() {
    $config = array(
      //some field specific config
    );

    $globalConfig = array(
      //som global config
    );

    parent::__construct($config, $globalConfig);
  }
}

```

Another purpose of separating a formatter into another class could be the re-use of certain configurations as you can always access `$this->config` or `$this->globalConfig` to access the parent configurations. 

### Nesting formatters

In addition to providing transformations for a given field it is also possible to use a nested base formatter to define transformations for a given field. This allows finer grained formatting for nested objects or arrays.
 
The example below shows you how to add a nested base formatter when the formatter is defined a seperate class. Note that its also possible to do everything inline without creating a seperate class. This is done in the exact same way as shown below.

```php

public function __construct() {

    $subFormatterConfig = array(
      array(
        'transformation' => [$this, 'unsetFields'],
        'params' => [['id', 'created', 'modified', 'locale']]
      ),
      array(
        'transformation' => [$this, 'transformFieldKeys'],
        'params' => ['canonicalized_email' => 'email']
      )
    );

    $config = array(
      'details' => array(
        'transformation' => new OutputFormatter([], $subFormatterConfig)
      )
    );
    
    parent::__construct($config);
  }
```

In the example above a new base formatter is added inline, but note that it is also possible to add a separate class instance. In fact, as long as the class extends OutputFormatter its possible to add any class as a sub formatter. The OutputFormatter is designed so that it can be used recursively as a sub-formatter. How many formatters you choose to nest is up to you. 
