<?php

namespace Tests\Bazookas\APIFrameworkBundle\Formatter\Base;

use Bazookas\APIFrameworkBundle\Formatter\Base\OutputFormatter;
use Symfony\Bundle\FrameworkBundle\Test\KernelTestCase;
use Symfony\Component\DependencyInjection\Container;

class OutputFormatterTest extends KernelTestCase
{

  /**
   * @var Container
   */
  private $container;

  public function setUp()
  {
    self::bootKernel();
    $this->container = self::$kernel->getContainer();
  }

  /**
   * Test converting the intput to an interger
   */
  public function testConvertToInteger() {
    $data = [
      'number' => '123',
      'number2' => '1234',
      'date' => new \DateTime()
    ];

    $config = [
      'number' => [
        'transformation' => [OutputFormatter::class, 'convertToType'],
        'params' => [OutputFormatter::TYPE_INTEGER]
      ],
      'number2' => [
        [
          'transformation' => [OutputFormatter::class, 'convertToType'],
          'params' => OutputFormatter::TYPE_INTEGER
        ]
      ]
    ];

    $formatter = new OutputFormatter($config);
    $data = $formatter->applyFieldTransformations($data, $config);

    $this->assertInternalType('int', $data['number']);
    $this->assertEquals($data['number'], 123);


    $this->assertInternalType('int', $data['number']);
    $this->assertEquals($data['number2'], 1234);

  }

  /**
   * Test if multiple field transformation are applied correctly.
   */
  public function testApplyMultipleFieldTransformations() {
    $data = [
      'number' => '123',
      'number2' => '1234',
      'date' => new \DateTime()
    ];

    $config = [
      'date' => [
        [
          'transformation' => 'date_format',
          'params' => 'U'
        ],
        [
          'transformation' => [OutputFormatter::class, 'convertToType'],
          'params' => OutputFormatter::TYPE_INTEGER
        ]
      ]
    ];

    $formatter = new OutputFormatter($config);
    $data = $formatter->applyFieldTransformations($data, $config);

    $this->assertInternalType('int', $data['date']);
  }

  /**
   * Test if a field transformation is applied correctly
   */
  public function testApplyFieldFormatter() {
    $data = [
      'number' => '123',
      'number2' => '1234',
      'user' => [
        'age' => 13.9,
        'username' => 'Testy'
      ],
      'ids' => [
        [
          'id' => 1
        ],
        [
          'id' => 2
        ],
        [
          'id' => 3
        ],
      ],
      'table' => [
        'id' => 3
      ],
      'empty' => [
        'name' => 'hi'
      ]
    ];

    $config = [
      'user' => [
        [
          'transformation' => new OutputFormatter([
            'age' => [
              'transformation' => 'ceil',
            ],
            //Test what happened when a field is included that's not in the data
            'id' => [
              'transformation' => 'floor'
            ]
          ])
        ]
      ],
      'ids' => [
        'transformation' => [OutputFormatter::class, 'propertyToScalarArray'],
      ],
      'table' => [
        'transformation' => [OutputFormatter::class, 'propertyToScalar']
      ],
      'empty' => [
        'transformation' => [OutputFormatter::class, 'propertyToScalar']
      ]
    ];

    $formatter = new OutputFormatter($config);
    $data = $formatter->applyFieldTransformations($data, $config);

    $this->assertEquals(14, $data['user']['age']);

    $this->assertInternalType('array', $data['ids']);

    $this->assertEquals(1, $data['ids'][0]);
    $this->assertEquals(2, $data['ids'][1]);
    $this->assertEquals(3, $data['ids'][2]);

    $this->assertEquals(3, $data['table']);

    $this->assertEquals(null, $data['empty']);

    //Assert unmodified
    $this->assertEquals('123', $data['number']);
  }

  /**
   * Test whether a global transformation is applied correctly
   */
  public function testApplyGlobalTransformation() {
    $data = [
      'number' => '123',
      'number2' => '1234',
      'date' => new \DateTime(),
      'user' => [
        'age' => 13.9,
        'username' => 'Testy',
        'details' => [
          'something' => 'hi'
        ]
      ]
    ];

    $config = [
      [
        'transformation' => [OutputFormatter::class, 'addToParent'],
        'params' => 'user'
      ]
    ];

    $formatter = new OutputFormatter([]);
    $data = $formatter->applyGlobalTransformations($data, $config);

    $this->assertTrue(isset($data['username']));

    //Assert unmodified
    $this->assertEquals('123', $data['number']);
  }

  /**
   * Test whether an array of global transformation is applied correctly
   */
  public function testApplicationOfMultipleGlobalTransformations() {
    $data = [
      'number' => '123',
      'number2' => '1234',
      'date' => new \DateTime(),
      'user' => [
        'age' => 13.9,
        'username' => 'Testy',
        'details' => [
          'something' => 'hi'
        ]
      ],
      'test' => [
        [
          'hi' => 3
        ]
      ]
    ];

    $config = [
      [
        [
          'transformation' => [OutputFormatter::class, 'addToParent'],
          'params' => 'user'
        ],
        [
          'transformation' => [OutputFormatter::class, 'unsetFields'],
          'params' => [['number2', 'date']]
        ],
        [
          'transformation' => [OutputFormatter::class, 'convertArrayToSingleton'],
          'params' => 'test'
        ]
      ]
    ];

    $formatter = new OutputFormatter([]);
    $data = $formatter->applyGlobalTransformations($data, $config);

    $this->assertTrue(isset($data['username']));
    $this->assertTrue(!isset($data['number2']));
    $this->assertTrue(!isset($data['date']));
    $this->assertEquals(3, $data['test']['hi']);

    //Assert unmodified
    $this->assertEquals('123', $data['number']);
  }

  /**
   * Test the format method
   */
  public function testFormat() {
    $data = [
      'number' => '123',
      'number2' => '1234',
      'date' => new \DateTime(),
      'user' => [
        'age' => 13.9,
        'username' => 'Testy',
        'details' => [
          'something' => 'hi'
        ]
      ],
      'test' => [
        [
          'hi' => 3
        ]
      ],
      'ids' => [
        [
          'id' => 1
        ],
        [
          'id' => 2
        ],
        [
          'id' => 3
        ],
      ],
      'table' => [
        'id' => 3
      ],
      'empty' => [
        'name' => 'hi'
      ]
    ];

    $config = [
      'user' => [
        [
          'transformation' => new OutputFormatter([
            'age' => [
              'transformation' => 'ceil',
            ],
            //Test what happend when a field is included that's not in the data
            'id' => [
              'transformation' => 'floor'
            ]
          ])
        ]
      ],
      'ids' => [
        'transformation' => [OutputFormatter::class, 'propertyToScalarArray'],
      ],
      'table' => [
        'transformation' => [OutputFormatter::class, 'propertyToScalar']
      ],
      'empty' => [
        'transformation' => [OutputFormatter::class, 'propertyToScalar']
      ]
    ];

    $globalConfig = [
      [
        [
          'transformation' => [OutputFormatter::class, 'addToParent'],
          'params' => 'user'
        ],
        [
          'transformation' => [OutputFormatter::class, 'unsetFields'],
          'params' => [['number2', 'date']]
        ],
        [
          'transformation' => [OutputFormatter::class, 'convertArrayToSingleton'],
          'params' => 'test'
        ]
      ]
    ];

    $formatter = new OutputFormatter($config, $globalConfig);
    $data = $formatter->format($data);

    //Assert field transformations
    $this->assertEquals(14, $data['age']);

    $this->assertInternalType('array', $data['ids']);

    $this->assertEquals(1, $data['ids'][0]);
    $this->assertEquals(2, $data['ids'][1]);
    $this->assertEquals(3, $data['ids'][2]);

    $this->assertEquals(3, $data['table']);

    $this->assertEquals(null, $data['empty']);

    //assert global transformations
    $this->assertTrue(isset($data['username']));
    $this->assertTrue(!isset($data['number2']));
    $this->assertTrue(!isset($data['date']));
    $this->assertEquals(3, $data['test']['hi']);

    //Assert unmodified
    $this->assertEquals('123', $data['number']);
  }

  /**
   * test global transformation 'usePropertyValueAsKey'
   */
  public function testValueAsPropertyTransformation() {
    $data = [
      [
        'test2' => 3,
        'type' => 'a_type'
      ]
    ];

    $globalConfig = [
      [
        'transformation' => [OutputFormatter::class, 'usePropertyValueAsKey'],
        'params' => ['type']
      ]
    ];

    $formatter = new OutputFormatter([]);
    $data = $formatter->applyGlobalTransformations($data, $globalConfig);

    $this->assertTrue(!isset($data['test2']));
    $this->assertTrue(isset($data['a_type']));
    $this->assertTrue(isset($data['a_type']['test2']));
  }

  /***
   * Test global transformation 'replaceKeyWithValue'
   */
  public function replaceKeyWithValueTransformation() {
    $data = [
      'test' => [
        'test2' => 3,
        'type' => 'a_type'
      ]
    ];

    $globalConfig = [
      [
        'transformation' => [OutputFormatter::class, 'replaceKeyWithValue'],
        'params' => [['test' => 'type']]
      ]
    ];

    $formatter = new OutputFormatter([]);
    $data = $formatter->applyGlobalTransformations($data, $globalConfig);

    $this->assertTrue(!isset($data['test']));
    $this->assertTrue(isset($data['a_type']));

  }

  /**
   * Test global transformation 'renameKey'
   */
  public function testRenameKeyTransformation() {
    $data = [
      'test' => 3
    ];

    $globalConfig = [
      [
        'transformation' => [OutputFormatter::class, 'renameKey'],
        'params' => [['test' => 'type']]
      ]
    ];

    $formatter = new OutputFormatter([]);
    $data = $formatter->applyGlobalTransformations($data, $globalConfig);

    $this->assertTrue(!isset($data['test']));
    $this->assertTrue(isset($data['type']));
    $this->assertEquals(3, $data['type']);
  }

  /**
   * test global transformation 'transformFieldKey'
   */
  public function testTransformFieldKeys() {
    $data = [
      'test' => 3,
      'test2' => 2
    ];

    $globalConfig = [
      [
        'transformation' => [OutputFormatter::class, 'transformFieldKey'],
        'params' => 'ucfirst'
      ]
    ];

    $formatter = new OutputFormatter([]);
    $data = $formatter->applyGlobalTransformations($data, $globalConfig);

    $this->assertTrue(isset($data['Test']));
    $this->assertTrue(!isset($data['test']));

    $this->assertTrue(isset($data['Test2']));
    $this->assertTrue(!isset($data['test2']));
  }

  /**
   * Test having a property containing an array of items
   */
  public function testArrayPropertyData() {
    $date1 = new \DateTime();
    $date2 = new \DateTime('now - 1 minute');
    $date3 = new \DateTime('now - 10 minute');

    $data = [
      'options' => [
        [
          'id' => 1,
          'name' => 'test',
          'created' => $date1
        ],
        [
          'id' => 2,
          'name' => 'test',
          'created' => $date2
        ],
        [
          'id' => 3,
          'name' => 'test',
          'created' => $date3
        ]
      ]
    ];

    $formatter = new OutputFormatter([
      'options' => [
        'transformation' => new OutputFormatter([
          'created' => [
            [
              'transformation' => 'date_format',
              'params' => 'U'
            ],
            [
              'transformation' => [OutputFormatter::class, 'convertToType'],
              'params' => OutputFormatter::TYPE_INTEGER
            ]
          ]
        ])
      ]
    ]);
    $formatter->format($data);

    $this->assertEquals($date1->getTimestamp(), $data['options'][0]['created']);
    $this->assertEquals($date2->getTimestamp(), $data['options'][1]['created']);
    $this->assertEquals($date3->getTimestamp(), $data['options'][2]['created']);
  }

  /**
   * Test passing array data to the formatter
   */
  public function testArrayData() {
    $date1 = new \DateTime();
    $date2 = new \DateTime('now - 1 minute');
    $date3 = new \DateTime('now - 10 minute');

    $data = [
      [
        'id' => 1,
        'name' => 'test',
        'created' => $date1
      ],
      [
        'id' => 2,
        'name' => 'test',
        'created' => $date2
      ],
      [
        'id' => 3,
        'name' => 'test',
        'created' => $date3
      ]
    ];

    $formatter = new OutputFormatter([
      'created' => [
        [
          'transformation' => 'date_format',
          'params' => 'U'
        ],
        [
          'transformation' => [OutputFormatter::class, 'convertToType'],
          'params' => OutputFormatter::TYPE_INTEGER
        ]
      ]
    ]);
    $formatter->format($data);

    $this->assertEquals($date1->getTimestamp(), $data[0]['created']);
    $this->assertEquals($date2->getTimestamp(), $data[1]['created']);
    $this->assertEquals($date3->getTimestamp(), $data[2]['created']);
  }

  /**
   * Test missing field transformation with an array of data
   */
  public function testMissingTransformationOnArrayData() {
    $date1 = new \DateTime();
    $date2 = new \DateTime('now - 1 minute');
    $date3 = new \DateTime('now - 10 minute');

    $data = [
      [
        'id' => 1,
        'name' => 'test',
        'created' => $date1
      ],
      [
        'id' => 2,
        'name' => 'test',
        'created' => $date2
      ],
      [
        'id' => 3,
        'name' => 'test',
        'created' => $date3
      ]
    ];

    $formatter = new OutputFormatter([
      'created' => [
        [
          'transformation' => 'date_format',
          'params' => 'U'
        ],
        [
          'transformation' => [OutputFormatter::class, 'convertToType'],
          'params' => OutputFormatter::TYPE_INTEGER
        ]
      ],
      'modified' => [
        [
          'transformation' => 'date_format',
          'params' => 'U'
        ],
        [
          'transformation' => [OutputFormatter::class, 'convertToType'],
          'params' => OutputFormatter::TYPE_INTEGER
        ]
      ]
    ]);
    $formatter->format($data);

    $this->assertEquals($date1->getTimestamp(), $data[0]['created']);
    $this->assertEquals($date2->getTimestamp(), $data[1]['created']);
    $this->assertEquals($date3->getTimestamp(), $data[2]['created']);
  }

  /**
   * Test global transformation 'addTranslationsToParent'
   */
//  public function testAddTranslationsToParent() {
//    $data = [
//      'id' => 2,
//      'details' => [
//        [
//          'name' => 'test',
//          'description' => 'a description',
//          'locale' => 'nl'
//        ],
//        [
//          'name' => 'test',
//          'description' => 'a description',
//          'locale' => 'fr'
//        ]
//      ]
//    ];
//
//    $formatter = new OutputFormatter([], [
//      [
//        'transformation' => [OutputFormatter::class, 'addTranslationsToParent'],
//        'params' => ['details']
//      ]
//    ]);
//    $formatter->format($data);
//
//    $this->assertTrue(isset($data['name']));
//    $this->assertTrue(isset($data['name']['nl']));
//    $this->assertTrue(isset($data['name']['fr']));
//
//    $this->assertTrue(isset($data['description']));
//    $this->assertTrue(isset($data['description']['nl']));
//    $this->assertTrue(isset($data['description']['fr']));
//  }

}
