<?php

namespace Bazookas\CommonBundle\Utility;

class HashUtility
{
  /**
   * Translates a number to a short alhanumeric version
   *
   * Translated any number up to 9007199254740992
   * to a shorter version in letters e.g.:
   * 9007199254740989 --> PpQXn7COf
   *
   * specifiying the second argument true, it will
   * translate back e.g.:
   * PpQXn7COf --> 9007199254740989
   *
   * this function is based on any2dec && dec2any by
   * fragmer[at]mail[dot]ru
   * see: http://nl3.php.net/manual/en/function.base-convert.php#52450
   *
   * If you want the alphaID to be at least 3 letter long, use the
   * $pad_up = 3 argument
   *
   * In most cases this is better than totally random ID generators
   * because this can easily avoid duplicate ID's.
   * For example if you correlate the alpha ID to an auto incrementing ID
   * in your database, you're done.
   *
   * The reverse is done because it makes it slightly more cryptic,
   * but it also makes it easier to spread lots of IDs in different
   * directories on your filesystem. Example:
   * $part1 = substr($alpha_id,0,1);
   * $part2 = substr($alpha_id,1,1);
   * $part3 = substr($alpha_id,2,strlen($alpha_id));
   * $destindir = "/".$part1."/".$part2."/".$part3;
   * // by reversing, directories are more evenly spread out. The
   * // first 26 directories already occupy 26 main levels
   *
   * @author  Kevin van Zonneveld <kevin@vanzonneveld.net>
   * @author  Simon Franz
   * @author  Deadfish
   * @author  SK83RJOSH
   * @copyright 2008 Kevin van Zonneveld (http://kevin.vanzonneveld.net)
   * @license   http://www.opensource.org/licenses/bsd-license.php New BSD Licence
   * @version   SVN: Release: $Id: alphaID.inc.php 344 2009-06-10 17:43:59Z kevin $
   * @link    http://kevin.vanzonneveld.net/
   *
   * @param mixed $in String or long input to translate
   * @param boolean $to_num Reverses translation when true
   * @param mixed $pad_up Number or boolean padds the result up to a specified length
   * @param string $salt Supplying a password makes it harder to calculate the original ID
   *
   * @param string $characters
   * @return mixed string or long
   */
  public static function alphaID($in, bool $to_num = false, $pad_up = false, string $salt = null, string $characters = 'abcdefghjklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ')
  {
    $base  = strlen($characters);

    if ($salt !== null) {
      // Although this function's purpose is to just make the
      // ID short - and not so much secure,
      // with this patch by Simon Franz (http://blog.snaky.org/)
      // you can optionally supply a password to make it harder
      // to calculate the corresponding numeric ID

      for ($n = 0; $n < strlen($characters); $n++) {
        $i[] = substr($characters, $n, 1);
      }

      $salt_hash = hash('sha256',$salt);
      $salt_hash = (strlen($salt_hash) < strlen($characters) ? hash('sha512', $salt) : $salt_hash);

      for ($n = 0; $n < strlen($characters); $n++) {
        $p[] =  substr($salt_hash, $n, 1);
      }

      array_multisort($p, SORT_DESC, $i);
      $characters = implode($i);
    }

    if ($to_num) {
      // Digital number  <<--  alphabet letter code
      $len = strlen($in) - 1;
      $out = 0;

      for ($t = $len; $t >= 0; $t--) {
        $bcp = bcpow($base, $len - $t);
        $out = $out + strpos($characters, substr($in, $t, 1)) * $bcp;
      }

      if (is_numeric($pad_up)) {
        $pad_up--;

        if ($pad_up > 0) {
          $out -= pow($base, $pad_up);
        }
      }
    } else {
      $out = '';

      // Digital number  -->>  alphabet letter code
      if (is_numeric($pad_up)) {
        $pad_up--;

        if ($pad_up > 0) {
          $in += pow($base, $pad_up);
        }
      }

      for ($t = ($in != 0 ? floor(log($in, $base)) : 0); $t >= 0; $t--) {
        $bcp = bcpow($base, $t);
        $a   = floor($in / $bcp) % $base;
        $out = $out . substr($characters, $a, 1);
        $in  = $in - ($a * $bcp);
      }
    }

    return $out;
  }
}
