## Importing files

### Getting started

Before diving into specifics, a list of steps you need to take:

- [create a service](#creating-the-import-service) handling the actual import
  - parses the data row-per-row
  - your service must handle persistance
- [create a page builder](#creating-the-page-builder) to configure your importer
- [edit your controller](#modifying-the-list-builder-controller) to
  - render the import form
  - use the import service you made earlier

### Usage

For best results use csv's exported from google docs. The import form does allow you to upload any type of csv, but you may run into parsing issues with this (but feel free to experiment with or improve on this component). 

#### Creating the import service

The list page builder uses a service `demo.example.import_service` to handle the actual file uploads and all of the importing logic. So create an import service by extending the `Bazookas\ExportBundle\Services\Import\Base\BaseImportService`. This service requires that you define three functions, as shown in the example below.

```php
<?php
namespace DemoBundle\Services;

use Bazookas\ExportBundle\Services\Import\Base\BaseImportService;
use DemoBundle\Entity\Example;

class ExampleImportService extends BaseImportService
{
  /**
   * @param array $rowData
   * @return array
   */
  protected function parseRow(array &$rowData)
  {
    $entity = new Example();

    $entity
      ->setPublished($rowData['Published'] === '1' ? true : false)
      ->setTitle($rowData['Title'])
    ;

    $this->entityManager->persist($entity);

    return $rowData;
  }

  /**
   * Either return a numerically indexed array containing the expected column headers, or an associative array
   * mapping the headers in the file to the header that should be used when creating the $rowData arrays
   * @return array
   */
  protected function getExpectedColumnHeaders()
  {
    return [
      'Published',
      'Title'
    ];
  }

  /**
   * This function is called when the entire document has been parsed.
   */
  protected function onComplete()
  {
    $this->entityManager->flush();
  }
}
```

This service expects two parameters, the entity manager and the project root directory:

```yml
  demo.example.import_service:
    class: DemoBundle\Services\ExampleImportService
    arguments:
      - "@doctrine.orm.entity_manager"
      - '%kernel.root_dir%/../'
```

For each row parsed by the service the `parseRow` function is invoked. It is up to you to do something with this parsed data. The `onComplete` function provides a callback to perform actions when the entire file has been parsed (such as persisting all of your entities to the database).

The `getExpectedColumnHeaders` function allows you to define an array of expected headers. If the imported file does not exactly match these headers, then the service will throw a new `ImportException` which can be handled using a `try-catch` as shown in the `modifyListBuilder` example above. The `getExpectedColumnHeaders` function allows you to define a numerically indexed array of headers: `['headerA', 'headerB']`. Alternatively you can also use an associative array:

```php
[
  'headerA' => 'parseRowIndexA',
  'headerB' => 'parseRowIndexB'
]
```

If you provide an associate array the `parseRow` function will receive data as an associative array:

```php
[
  'parseRowIndexA' => 0,
  'parseRowIndexB' => 1
]
```

If you did not specify an associative array you will receive a normal numerically indexed array instead.


#### Creating the page builder

Let's assume you want to add an import button to a basic list page. Start by creating a custom page builder:

```php
<?php
namespace DemoBundle\PageBuilder;

use Bazookas\AdminBundle\AdminElements\Containers\CollapsibleContainer;
use Bazookas\AdminBundle\AdminElements\Elements\FormPanelElement;
use Bazookas\AdminBundle\PageBuilder\ListPageBuilder;
use Symfony\Component\Form\Form;
use Symfony\Component\HttpFoundation\Request;

class ExamplePageBuilder extends ListPageBuilder
{
  /**
   * @var Form
   */
  protected $form;

  /**
   * @return Form
   */
  public function getForm()
  {
    return $this->form;
  }

  /**
   * @param Form $form
   * @return $this
   */
  public function setForm(Form $form)
  {
    $this->form = $form;

    return $this;
  }

  /**
   * @inheritDoc
   */
  public function buildPage(Request $request) {
    //Add a header element
    $this->elements[] = $this->generateHeaderElement();

    $this->elements[] = $this->buildImportForm();

    //Add the list table
    $this->elements[] = $this->buildListTable($request);
  }

  /**
   * @return CollapsibleContainer
   */
  private function buildImportForm()
  {
    $formElement = new FormPanelElement([], [
      'form' => $this->getForm()->createView(),
      'label' => 'admin.entities.example.addLabel',
    ]);

    $form = $this->getForm();

    return new CollapsibleContainer([$formElement], [
      'id' => 'js-inline-add-form',
      'initiallyCollapsed' => !(!$form->isValid() && $form->isSubmitted()),
    ]);
  }
}
```

All this page builder does is add a new collapsible import file form to the top of your list table page.

#### Modifying the list builder (controller)

In your `modifyListBuilder` function add a new page action and the logic necessary to handle the import form:

```php
<?php
namespace DemoBundle\Controller;

use Bazookas\AdminBundle\AdminElements\Elements\Actions\PageActions\ImportPageActionElement;
use Bazookas\AdminBundle\Controller\Base\BaseAdminListController;
use Bazookas\AdminBundle\PageBuilder\Interfaces\ListPageBuilderInterface;
use Bazookas\ExportBundle\Entity\GenericFileEntity;
use Bazookas\ExportBundle\Exception\ImportException;
use Bazookas\ExportBundle\Form\ImportFileForm;
use DemoBundle\PageBuilder\ExamplePageBuilder;
use Symfony\Component\HttpFoundation\Request;

class ExampleAdminController extends BaseAdminListController
{
  /**
   * Don't forget to add the custom page builder to the builders list in the constructor.
   **/
  function __construct()
  {
    // make sure parent construct is called!
    parent::__construct();
    
    $this->builders[self::ACTION_LIST] = ExamplePageBuilder::class;
    // ... other builders ...
  }
  
  // {... other [required] controller methods ...}

  /**
   * @param Request $request
   * @param ListPageBuilderInterface $builder
   * @return ListPageBuilderInterface
   */
  protected function modifyListBuilder(Request $request, ListPageBuilderInterface $builder)
  {
    // ... other modify functionality ...
  
    // Add the import functionality
    $entity = new GenericFileEntity();
    $form = $this->createForm(ImportFileForm::class, $entity);
  
    $form->handleRequest($request);
    $builder->setForm($form);
  
    if ($form->isValid() && $form->isSubmitted()) {
      $service = $this->get('demo.example.import_service');
      $entity = $service->handleUpload($entity);
      try {
        $service->process($entity, true, true);
  
        $translator = $this->get('translator');
  
        $this->addFlash('success', $translator->trans('admin.entities.example.onImported', [
          '%filename%' => $entity->getOriginalFileName()
        ], 'admin'));
      } catch(ImportException $e) {
        $this->addFlash('error', $e->getMessage());
      }
    }
  
    // Add the import button
    $importPageAction = new ImportPageActionElement([], [
      'label' => 'admin.entities.example.import',
      'action' => self::ACTION_LIST,
      'route' => $request->get('_route'),
      'attr' => [
        'data-toggle' => 'collapse',
        'role' => 'button',
        'aria-expanded' => !$form->isValid() && $form->isSubmitted() ? 'true' : 'false',
        'aria-controls' => 'js-inline-add-form',
        'href' => '#js-inline-add-form'
      ]
    ]);
  
    $builder->prependAction($importPageAction);
  
    return $builder;
  }  
}
```
