<?php

namespace Bazookas\ExportBundle\Services;

use Bazookas\CronBundle\Entity\Cron;
use Bazookas\ExportBundle\Services\Interfaces\DataSelectorInterface;
use Bazookas\ExportBundle\Services\Interfaces\ExportServiceInterface;
use Doctrine\ORM\QueryBuilder;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RedirectResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\OptionsResolver\OptionsResolver;

class ExportService implements ExportServiceInterface {

  private $container;

  const TYPE_CSV = 'csv';
  const TYPE_XLSX = 'xlsx';

  public function __construct(ContainerInterface $container) {
    $this->container = $container;
  }

  /**
   * @param array $config
   * @return array
   */
  protected function resolveConfig(array $config) {
    $configResolver = new OptionsResolver();
    $configResolver
      ->setDefined([
        'sheets',
        'extraParams',
        'fileName',
        'format',
        'dataService',
        'renderSync',
        'email',
        'fileCreationService'
      ])
      ->setRequired([
        'sheets',
        'fileName'
      ])
      ->setDefaults([
        'dataService' => 'bazookas.export.service.DataSelectorService',
        'renderSync' => true,
        'format' => self::TYPE_CSV,
        'extraParams' => [],
        'fileCreationService' => 'bazookas.export.service.ExportFileCreationService'
      ])
      ->setAllowedTypes('sheets', 'array')
      ->setAllowedTypes('extraParams', 'array')
      ->setAllowedTypes('fileName', 'string')
      ->setAllowedTypes('dataService', 'string')
      ->setAllowedTypes('renderSync', 'bool')
      ->setAllowedTypes('email', 'string')
    ;

    $sheetResolver = new OptionsResolver();
    $sheetResolver
      ->setDefined([
        'entity',
        'fields',
        'sheetName',
        'queryBuilder'
      ])

      ->setDefaults([
        'conditions' => [],
        'parameters' => [],
        'sheetName' => 'sheet 1'
      ])

      ->setAllowedTypes('entity', 'string')
      ->setAllowedTypes('fields', 'array')
      ->setAllowedTypes('parameters', 'array')
      ->setAllowedTypes('sheetName', 'string')
      ->setAllowedTypes('queryBuilder', [QueryBuilder::class])

      ->setAllowedValues('entity', function($value) {
        return class_exists($value);
      })

      ->setRequired([
        'entity',
        'fields',
      ])
    ;

    $resolvedConfig = $configResolver->resolve($config);

    foreach($resolvedConfig['sheets'] as $key => $sheet) {
      $resolvedConfig['sheets'][$key] = $sheetResolver->resolve($sheet);
    }

    return $resolvedConfig;
  }

  /**
   * @param Request $request
   * @param array $exportConfig
   * @return RedirectResponse|\Symfony\Component\HttpFoundation\Response
   * @throws \Exception
   */
  public function export(Request $request, array $exportConfig) {
    //resolve the configuration
    $exportConfig = $this->resolveConfig($exportConfig);

    if (!$exportConfig['renderSync']) {
      $this->renderAsync($exportConfig);
      return $this->getRedirectResponse($request);
    } else {
      return $this->renderSync($exportConfig);
    }
  }

  /**
   * @param Request $request
   * @return RedirectResponse
   */
  protected function getRedirectResponse(Request $request) {
    $referrer = $request->headers->get('referer');

    if (empty($referrer)) {
      $router = $this->container->get('router');
      $referrer = $router->generate('bazookas_admin_homepage');
    }

    return new RedirectResponse($referrer);
  }

  /**
   * @param array $exportConfig
   */
  protected function renderAsync(array $exportConfig) {
    $uniqueFileName = uniqid($exportConfig['fileName'] . '_', true) . $exportConfig['format'];
    $exportConfig['uniqueFileName'] = $uniqueFileName;
    $configContent = serialize($exportConfig);
    $tempDir = $this->container->getParameter('kernel.root_dir') . '/../var/tmp/';
    $tempDir = realpath($tempDir);

    $tempFile = tempnam($tempDir, 'tmp');
    $handle = fopen($tempFile, "w");
    fwrite($handle, $configContent);
    fclose($handle);
    chmod($tempFile, 0777);

    $root = (!empty($_SERVER['HTTPS']) ? 'https' : 'http') . '://' . $_SERVER['HTTP_HOST'] . '/';

    $cronCommand = new Cron();
    $cronCommand
      ->setCommand('bin/console export:create ' . $tempFile)
      ->setExecutionInterval(1)
      ->setIntervalUnit(Cron::INTERVAL_HOURS)
      ->setNextRunOn(new \DateTime())
      ->setMaxExecutionTime(300)
      ->setMaxExecutionsBeforeFail(1)
      ->setRemoveAfterExecute(true)
      ->setCronService('bazookas.export.service.ExportCronService')
      ->setEmail($exportConfig['user'])
      ->setEmailParameters([
        'filePath' => $root . $uniqueFileName,
      ])
    ;

    $em = $this->container->get('doctrine.orm.entity_manager');
    $em->persist($cronCommand);
    $em->flush();

    $this->container->get('session')->getFlashBag()->add('notice', 'An email with the download link will be sent to ' . $exportConfig['user']);
  }

  /**
   * @param array $exportConfig
   * @return \Symfony\Component\HttpFoundation\Response
   * @throws \Exception
   */
  protected function renderSync(array $exportConfig)
  {
    //retrieve the service
    $dataService = $this->container->get($exportConfig['dataService']);
    if (!$dataService instanceof DataSelectorInterface) {
      throw new \Exception('DataService does not implement ' . DataSelectorInterface::class);
    }

    $exportData = $dataService->getData($exportConfig);
    $fileCreationService = $this->container->get('bazookas.export.service.ExportFileCreationService');

    //The file function will always be 'create' . <format> . 'File'; For example 'createCsvFile'
    $fileFunction = 'create' . ucfirst($exportConfig['format']) . 'File';

    if (method_exists($fileCreationService, $fileFunction)) {
      return $fileCreationService->$fileFunction($exportData, $exportConfig);
    } else {
      throw new \Exception('Unsupported file format \'' . $exportConfig['format'] . '\' no createFile function found in service \'' . get_class($fileCreationService) . '\'');
    }
  }
}
